# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals
import pytest
import numpy as np
from os import path

import IIRrational
from wield.iirrational import v2
from wield.iirrational.testing.plots import plot_on_assert

from wield.iirrational.testing.utilities import (
    sign_validate_and_plot_hint,
    stability_validate_and_plot_hint,
)

try:
    from IIRrational_test_data import IIRrational_data_dev
    from IIRrational_test_data import matlab as matlab_test_data
except ImportError:
    module_import_skip = True
else:
    module_import_skip = False

localpath = path.split(__file__)[0]
skipdeco = pytest.mark.skipif(
    module_import_skip, reason="cannot import IIRrational_test_data"
)


@skipdeco
def test_alphas1(request, browser, plotsections, plot_verbosity):
    fname = matlab_test_data.matfiles["alphas_fdomain.txt"]
    alphas = np.loadtxt(fname)
    # assemble back as complex numbers (each column of a is a TF)
    fr = alphas[:, 0]
    a = np.zeros((alphas.shape[0], (alphas.shape[1] - 1) // 2), dtype=complex)
    for i in range(a.shape[1]):
        a[:, i] = alphas[:, i + 1] + 1j * alphas[:, i + 2]

    idx = np.where(np.logical_and(fr > 30, fr < 300))[0]

    fit = wield.iirrational.v2.data2filter(
        data=a[idx, 0],
        F_Hz=fr[idx],
        SNR=100,
        # mode = 'rational',
        mode="chebydebug",
        F_nyquist_Hz=None,
        log_level=10,
        baseline_only=True,
        # poles_overlay = (-1,),
        # zeros_overlay = (-30,),
        # never_unstable_poles = True,
        # never_unstable_zeros = True,
        log_level_debug=10,
    )
    with plot_on_assert(__file__, request, fit.fitter, plot_anyway=True):
        pass
    return


@skipdeco
def test_alphas1_v1good(request, browser, plotsections, plot_verbosity):
    fname = matlab_test_data.matfiles["alphas_fdomain.txt"]
    alphas = np.loadtxt(fname)
    # assemble back as complex numbers (each column of a is a TF)
    fr = alphas[:, 0]
    a = np.zeros((alphas.shape[0], (alphas.shape[1] - 1) // 2), dtype=complex)
    for i in range(a.shape[1]):
        a[:, i] = alphas[:, i + 1] + 1j * alphas[:, i + 2]

    idx = np.where(np.logical_and(fr > 30, fr < 300))[0]

    fit = wield.iirrational.v1.data2filter(
        data=a[idx, 0],
        F_Hz=fr[idx],
        SNR=100,
        order_initial=50,
        F_nyquist_Hz=None,
    )
    with plot_on_assert(__file__, request, fit.fitter, plot_anyway=True):
        pass
    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


from ..codings_cmn import (
    CodingTypeZ,
)


class CodingRealnlBW(CodingTypeZ):
    N_parameters = 1
    unstable = False
    negative_real = False
    p_nl_BW = 0
    min_BW_Hz = 0

    def option_set(self, min_BW_Hz=None, **kwargs):
        super(CodingRealnlBW, self).option_set(**kwargs)
        if min_BW_Hz is not None:
            self.min_BW_Hz = min_BW_Hz
        # TODO, should modify p_nl_A for the new max_amplitude
        return

    def check_dist_limit(self, thresh=1):
        if self.sys.distance_limit_auto >= thresh:
            self.min_BW_Hz = self.sys.distance_limit(0)
            return True
        return False

    def update(self, nl_BW):
        self.p_nl_BW = nl_BW

    def reduce(self):
        return [self.p_nl_BW]

    def update_roots(self, r1):
        assert r1.imag == 0
        amp = r1
        self.check_dist_limit(thresh=1)

        if amp < 0:
            self.negative_real = True
            amp = -amp
        else:
            self.negative_real = False

        BW_Hz = (amp - 1) * self.sys.F_nyquist_Hz

        if BW_Hz <= 0:
            self.unstable = False
            BWp = -BW_Hz
        else:
            self.unstable = True
            BWp = BW_Hz
        if BWp > 2 * self.min_BW_Hz:
            self.p_nl_BW = (BWp - self.min_BW_Hz) ** 0.5
            return True
        else:
            self.p_nl_BW = self.min_BW_Hz ** 0.5
            return False

    @property
    def gain_effect(self):
        if self.unstable:
            return -1
        else:
            return 1

    @property
    def BW(self):
        if not self.unstable:
            BW = -self.p_nl_BW ** 2
        else:
            BW = self.p_nl_BW ** 2
        return BW

    @property
    def F_Hz(self):
        return 0

    def transfer(self):
        self.check_dist_limit(thresh=2)
        if not self.negative_real:
            amp = 1 + self.BW / self.sys.F_nyquist_Hz
        else:
            amp = -(1 + self.BW / self.sys.F_nyquist_Hz)
        # real, log amplitude, positive
        return 1 - amp * self.sys.Xzn_grid

    def derivative(self):
        self.check_dist_limit(thresh=2)
        if self.disable:
            return []
        # real/imaginary part of root
        if not self.unstable:
            DampDl = -(2 * self.p_nl_BW) / self.sys.F_nyquist_Hz
            amp = 1 - self.p_nl_BW / self.sys.F_nyquist_Hz
        else:
            DampDl = +(2 * self.p_nl_BW) / self.sys.F_nyquist_Hz
            amp = 1 + self.p_nl_BW / self.sys.F_nyquist_Hz

        if self.negative_real:
            DampDl *= -1
            amp *= -1

        return [
            -DampDl * self.sys.Xzn_grid,
        ]

    def roots_r(self):
        if not self.negative_real:
            amp = 1 + self.BW / self.sys.F_nyquist_Hz
        else:
            amp = -(1 + self.BW / self.sys.F_nyquist_Hz)
        return [amp]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
Utilities to manipulate ZPK roots S to/from Z, and make transfer functions
"""

import numpy as np
import warnings

from .roots_bin import roots_bin_type

imag_tol = 1e-12


def TF_ZPK(F_Hz, ZPK, F_nyquist_Hz=None, group_len=1):
    """
    Stably generates the ZPK transfer function, interlacing zeros and poles to prevent small number accumulating
    """
    # TODO deprecate
    Z, P, K = ZPK
    if F_nyquist_Hz is None:
        X = 1j * F_Hz
    else:
        X = np.exp(1j * np.pi * F_Hz / F_nyquist_Hz)

    Z = np.asarray(Z)
    P = np.asarray(P)
    mlen = max(len(P), len(Z))
    h = 1
    for idx in range((mlen - 1) // group_len + 1):
        zs = Z[idx * group_len : (idx + 1) * group_len]
        ps = P[idx * group_len : (idx + 1) * group_len]
        h *= np.polynomial.polynomial.polyvalfromroots(X, zs)
        h /= np.polynomial.polynomial.polyvalfromroots(X, ps)
    return K * h


def ZtoS(ZPK, F_nyquist_Hz, F_Hz=None, gain_scale_policy="median", error=False):
    # use a list for mutability due to local/nonlocal referencing semantics
    gain_flip = [1]

    def roots_move(roots_list):
        moved = []
        for r in roots_list:
            if abs(r.imag) < imag_tol:
                if r < 0:
                    if error:
                        raise RuntimeError("Negative real root found in ZtoS")
                    else:
                        warnings.warn("Negative real root found in ZtoS, dropping")
                    gain_flip[0] *= -1
                else:
                    real_Hz = -(1 - r) * F_nyquist_Hz / np.pi
                    moved.append(real_Hz)
            else:
                F_Hz = np.angle(r) / np.pi * F_nyquist_Hz
                real_Hz = -(1 - abs(r)) * F_nyquist_Hz / np.pi
                moved.append(real_Hz + 1j * F_Hz)
        return moved

    Z_z, P_z, K_z = ZPK
    Z_s = roots_move(Z_z)
    P_s = roots_move(P_z)

    if F_Hz is None:
        gain_scale = 1
    else:
        F_Hz = np.asarray(F_Hz)
        xfer_ratio = abs(
            TF_ZPK(F_Hz, (Z_s, P_s, 1), F_nyquist_Hz=None)
            / TF_ZPK(F_Hz, (Z_z, P_z, 1), F_nyquist_Hz=F_nyquist_Hz)
        )
        if gain_scale_policy == "median":
            gain_scale = np.median(xfer_ratio)
        elif gain_scale_policy in ["mean", "average"]:
            gain_scale = np.mean(xfer_ratio)
        else:
            raise RuntimeError(
                'Unrecognized setting for gain_scale_policy, must be "median", "average" (or "Mean").'
            )
    K_s = K_z * gain_flip[0] / gain_scale
    return (np.asarray(Z_s), np.asarray(P_s), K_s)


def StoZ(ZPK, F_nyquist_Hz, F_Hz=None, gain_scale_policy="median", error=False):
    # use a list for mutability due to local/nonlocal referencing semantics
    gain_flip = [1]

    def roots_move(roots_list):
        moved = []
        for r in roots_list:
            if abs(r.imag / r.real) < imag_tol:
                real_D = 1 + r * np.pi / F_nyquist_Hz
                moved.append(real_D)
            else:
                if r.imag >= F_nyquist_Hz:
                    if error:
                        raise RuntimeError(
                            "Root Frequency above the nyquist encountered in StoZ"
                        )
                    else:
                        warnings.warn(
                            "Root Frequency above the nyquist encountered in StoZ, dropping"
                        )
                else:
                    F_D = np.exp(1j * np.pi * r.imag / F_nyquist_Hz)
                    amp_D = 1 + r.real * np.pi / F_nyquist_Hz
                    moved.append(amp_D * F_D)
        return moved

    Z_s, P_s, K_s = ZPK
    Z_z = roots_move(Z_s)
    P_z = roots_move(P_s)
    if F_Hz is None:
        gain_scale = 1
    else:
        F_Hz = np.asarray(F_Hz)
        xfer_ratio = abs(
            TF_ZPK(F_Hz, (Z_z, P_z, 1), F_nyquist_Hz=F_nyquist_Hz)
            / TF_ZPK(F_Hz, (Z_s, P_s, 1), F_nyquist_Hz=None)
        )

        if gain_scale_policy == "median":
            gain_scale = np.median(xfer_ratio)
        elif gain_scale_policy in ["mean", "average"]:
            gain_scale = np.mean(xfer_ratio)
        else:
            raise RuntimeError(
                'Unrecognized setting for gain_scale_policy, must be "median", "average" (or "Mean").'
            )

    K_z = K_s * gain_flip[0] / gain_scale
    return (Z_z, P_z, K_z)


def SorZtoSorZ(
    ZPK,
    F_nyquist_Hz_in,
    F_nyquist_Hz_out,
    F_Hz=None,
    gain_scale_policy="median",
    error=False,
):
    if F_nyquist_Hz_in is not None:
        ZPK_S = ZtoS(
            ZPK,
            F_nyquist_Hz=F_nyquist_Hz_in,
            F_Hz=F_Hz,
            gain_scale_policy=gain_scale_policy,
            error=error,
        )
    else:
        ZPK_S = ZPK

    if F_nyquist_Hz_out is not None:
        ZPK_2 = StoZ(
            ZPK_S,
            F_nyquist_Hz=F_nyquist_Hz_out,
            F_Hz=F_Hz,
            gain_scale_policy=gain_scale_policy,
            error=error,
        )
    else:
        ZPK_2 = ZPK_S
    return ZPK_2


def ZPK_fill(
    ZPK=None,
    Z=None,
    P=None,
    K=None,
    F_nyquist_Hz=None,
):
    if Z is None:
        Z = ZPK[0]
    if P is None:
        P = ZPK[1]
    if K is None:
        K = ZPK[2]

    z_r, z_c, z_u, pol = roots_bin_type(
        Z,
        policy="auto",
        F_nyquist_Hz=F_nyquist_Hz,
    )
    if len(z_u) > 0:
        raise RuntimeError("Found unmatched zeros: {0}".format(str(z_u)))
    zeros = []
    zeros.extend(z_r)
    zeros.extend(z_c)
    zeros.extend([r.conjugate() for r in z_c])

    p_r, p_c, p_u, pol = roots_bin_type(
        P,
        policy="auto",
        F_nyquist_Hz=F_nyquist_Hz,
    )
    if len(p_u) > 0:
        raise RuntimeError("Found unmatched poles: {0}".format(str(p_u)))

    poles = []
    poles.extend(p_r)
    poles.extend(p_c)
    poles.extend([r.conjugate() for r in p_c])
    return [zeros, poles, K]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""
import sys
import numpy as np

from ..utilities import np as util_np
from .. import fitters_ZPK
from .. import representations
from ..utilities import np as np_utils

from . import arguments
from .arguments import (
    grab_kwargs,
    grab_kwarg_hints,
)
from . import order_reduce
from . import algorithms
from . import phase_patch
from . import fit_aid
from . import results_aid_adv
from . import rational_fits
from . import SNR_adjustments
from . import autogen_docstr
from . import AAA_fits
from .algorithms import order_reduce_flip


def data2filter(*args, **kw):
    """
    """
    kwput = dict()
    kw, ZPKrep = arguments.kw_ZPKrep_build(args, kw)

    # pop all advanced hints
    hints = kw.pop("hints", {})
    aid = fit_aid.FitAid(hints=hints)

    # setup logging early so that error reporting preferences are done before any
    # errors are reported
    grab_kwarg_hints(aid, kw, arguments.logging.kw_hints, kwput=kwput)

    tee_logfile = aid.hint('tee_logfile')
    print("TEE_LOGFILE", tee_logfile)
    if tee_logfile is not None:
        # open it with 'w' just to clear it
        with open(tee_logfile, 'w') as F:
            pass

    helparg = grab_kwargs(aid, kw, arguments.kw_hints, "help")


    if helparg:
        raise NotImplementedError("Can't provide help functionality yet")

    # now pull the defaults from chain
    chain = grab_kwargs(aid, kw, arguments.kw_hints, "chain")
    if chain is not None:
        raise NotImplementedError("WIP")

    grab_kwarg_hints(aid, kw, arguments.standardargs.kw_hints_pre, kwput=kwput)
    data = grab_kwargs(aid, kw, arguments.kw_hints, "data")
    F_Hz = grab_kwargs(aid, kw, arguments.kw_hints, "F_Hz")
    SNR = grab_kwargs(aid, kw, arguments.kw_hints, "SNR")
    COH = grab_kwargs(aid, kw, arguments.kw_hints, "coherence")
    SNR_phase_rel = grab_kwargs(aid, kw, arguments.kw_hints, "SNR_phase_relative")
    emphasis = grab_kwargs(aid, kw, arguments.kw_hints, "emphasis")
    select = grab_kwargs(aid, kw, arguments.kw_hints, "select")
    F_nyquist_Hz = grab_kwargs(aid, kw, arguments.kw_hints, "F_nyquist_Hz", kwput=kwput)

    mode = grab_kwargs(aid, kw, arguments.kw_hints, "mode", kwput=kwput)

    ZPK = grab_kwargs(aid, kw, arguments.kw_hints, "ZPK", kwput=kwput)
    zeros = grab_kwargs(aid, kw, arguments.kw_hints, "zeros", kwput=kwput)
    poles = grab_kwargs(aid, kw, arguments.kw_hints, "poles", kwput=kwput)
    gain = grab_kwargs(aid, kw, arguments.kw_hints, "gain", kwput=kwput)

    if zeros is None and ZPK is not None:
        zeros = ZPK.zeros
    if zeros is None and ZPKrep is not None:
        zeros = ZPKrep.zeros
    if zeros is None:
        zeros = ()

    if poles is None and ZPK is not None:
        poles = ZPK.poles
    if poles is None and ZPKrep is not None:
        poles = ZPKrep.poles
    if poles is None:
        poles = ()

    if gain is None and ZPK is not None:
        gain = ZPK.gain
    if gain is None and ZPKrep is not None:
        gain = ZPKrep.gain
    if gain is None:
        gain = 1

    ZPK_overlay = grab_kwargs(aid, kw, arguments.kw_hints, "ZPK_overlay", kwput=kwput)
    zeros_overlay = grab_kwargs(
        aid, kw, arguments.kw_hints, "zeros_overlay", kwput=kwput
    )
    poles_overlay = grab_kwargs(
        aid, kw, arguments.kw_hints, "poles_overlay", kwput=kwput
    )

    if zeros_overlay is None and ZPK_overlay is not None:
        zeros_overlay = ZPK_overlay.zeros
    if zeros_overlay is None and ZPKrep is not None:
        zeros_overlay = ZPKrep.zeros_overlay
    if zeros_overlay is None:
        zeros_overlay = ()

    if poles_overlay is None and ZPK_overlay is not None:
        poles_overlay = ZPK_overlay.poles
    if poles_overlay is None and ZPKrep is not None:
        poles_overlay = ZPKrep.poles_overlay
    if poles_overlay is None:
        poles_overlay = ()

    if F_nyquist_Hz is not None:
        raise NotImplementedError("V2 currently only works in the Sf domain")

    if F_Hz is None and ZPKrep is not None:
        F_Hz = ZPKrep.F_Hz
    if data is None and ZPKrep is not None:
        data = ZPKrep.data
    if SNR is None and ZPKrep is not None:
        SNR = ZPKrep.W

    if SNR_phase_rel is None and ZPKrep is not None:
        SNR_phase_rel = ZPKrep.residuals_log_im_scale

    if SNR is not None:
        if isinstance(SNR, tuple):
            SNR = tuple(np_utils.broadcast_arrays_none(*SNR))
        else:
            SNR = (SNR,)

    if COH is not None:
        if isinstance(COH, tuple):
            COH = tuple(np_utils.broadcast_arrays_none(*COH))
        else:
            COH = (COH,)

        SNRs = []
        for coh in COH:
            SNRs.append((1 / (1 - coh ** 0.5) - 1) ** 0.5)
        if SNR is None:
            SNR = tuple(SNRs)
        else:
            SNR = SNR + tuple(SNRs)

    # now combine all of the SNR arrays into a single one
    if SNR is not None:
        iSNR = 0
        for snr in SNR:
            iSNR = iSNR + snr ** -2
        SNR = iSNR ** -0.5

    if SNR_phase_rel is None:
        SNR_phase_rel = 1

    F_Hz, data, SNR, SNR_phase_rel, emphasis = np_utils.broadcast_arrays_none(
        F_Hz, data, SNR, SNR_phase_rel, emphasis
    )
    F_Hz, data, SNR, SNR_phase_rel, emphasis = np_utils.select_through_none(
        select, F_Hz, data, SNR, SNR_phase_rel, emphasis
    )

    SNR_min = grab_kwargs(
        aid, kw, arguments.adjustments.kw_hints, "SNR_min", kwput=kwput
    )
    SNR_est_width = grab_kwargs(
        aid, kw, arguments.kw_hints, "SNR_estimate_width", kwput=kwput
    )
    if SNR is None:
        if SNR_est_width is None or SNR_est_width == 0:
            aid.log_warn(
                3,
                (
                    """
                SNR not specified, using unweighted "1". If data has many points
                with low SNR, the fit will be poor. Try 'SNR_est_width' to attempt
                an estimate using sample variance.
                """
                ).format(),
            )
            SNR = np.ones_like(F_Hz)
        else:
            aid.log_warn(
                3,
                (
                    """
                Estimating SNR from sample variance with nearby points (SNR_est_width={} > 0).
                This technique works semi-OK, but could probably be much better..
                use the resulting fit to estimate the sample variance and generate
                improved SNR estimates, iterate.
                """
                ).format(SNR_est_width),
            )
            from . import SNR_estimate

            SNR = SNR_estimate.SNR_estimate(F_Hz, data, width=SNR_est_width)
            if SNR_min is None:
                SNR_min = 1
            W_select = SNR > SNR_min
            if not np.all(W_select):
                F_Hz, data, SNR, SNR_phase_rel, emphasis = np_utils.select_through_none(
                    W_select, F_Hz, data, SNR, SNR_phase_rel, emphasis
                )
                aid.log_warn(
                    3,
                    (
                        """{} SNR<{} element(s) dropped (of {}).
                    Too many low SNR elements confuses the rational nonparametric fitter.
                    """
                    ).format(np.count_nonzero(~W_select), SNR_min, len(W_select)),
                )
    else:
        if SNR_min is None:
            SNR_min = 0
        W_select = SNR > SNR_min
        if not np.all(W_select):
            F_Hz, data, SNR, SNR_phase_rel, emphasis = np_utils.select_through_none(
                W_select, F_Hz, data, SNR, SNR_phase_rel, emphasis
            )
            aid.log_warn(
                3,
                "{} SNR=0 element(s) dropped (of {})".format(
                    np.count_nonzero(~W_select), len(W_select)
                ),
            )

    F_max_Hz = grab_kwargs(
        aid, kw, arguments.adjustments.kw_hints, "F_max_Hz", kwput=kwput
    )
    if F_max_Hz is not None:
        select_Fmax = F_Hz <= F_max_Hz
        aid.log_warn(
            3,
            "{} element(s) dropped (of {}) due to F_max_Hz={}".format(
                np.count_nonzero(~select_Fmax), len(select_Fmax), F_max_Hz
            ),
        )
        F_Hz, data, SNR, SNR_phase_rel, emphasis = np_utils.select_through_none(
            select_Fmax, F_Hz, data, SNR, SNR_phase_rel, emphasis
        )

    F_min_Hz = grab_kwargs(
        aid, kw, arguments.adjustments.kw_hints, "F_min_Hz", kwput=kwput
    )
    if F_min_Hz is not None:
        select_Fmax = F_Hz >= F_min_Hz
        aid.log_warn(
            3,
            "{} element(s) dropped (of {}) due to F_min_Hz={}".format(
                np.count_nonzero(~select_Fmax), len(select_Fmax), F_min_Hz
            ),
        )
        F_Hz, data, SNR, SNR_phase_rel, emphasis = np_utils.select_through_none(
            select_Fmax, F_Hz, data, SNR, SNR_phase_rel, emphasis
        )

    nan_select = np.isfinite(data)
    if not np.all(nan_select):
        F_Hz, data, SNR, SNR_phase_rel, emphasis = np_utils.select_through_none(
            nan_select, F_Hz, data, SNR, SNR_phase_rel, emphasis
        )
        aid.log_warn(3, "NaN's and/or inf's present in dataset. Dropping them.")

    ###################################### downsampling #######################
    downsampling_N = grab_kwargs(
        aid, kw, arguments.adjustments.kw_hints, "downsample", kwput=kwput
    )
    downsampling_type = grab_kwargs(
        aid, kw, arguments.adjustments.kw_hints, "downsample_type", kwput=kwput
    )
    downsampling_type = downsampling_type.lower()
    if downsampling_N is not None and downsampling_N < len(F_Hz):
        if downsampling_type == "log":
            F_groups = util_np.logspaced(F_Hz[0], F_Hz[-1], downsampling_N)
        if downsampling_type in ["linear", "lin"]:
            F_groups = np.linspace(F_Hz[0], F_Hz[-1], downsampling_N)
        if downsampling_type in ["loglin", "linlog"]:
            F_groups = util_np.loglinspaced(F_Hz[0], F_Hz[-1], downsampling_N)

        idx_groups = np.searchsorted(F_Hz, F_groups)
        idx_pairs = list(zip(idx_groups[:-1], idx_groups[1:]))

        lemph = []
        lF_Hz = []
        ldata = []
        lSNR = []
        lSNRpr = []
        for idx1, idx2 in idx_pairs:
            if idx1 == idx2:
                continue
            ldata.append(
                np.sum(data[idx1:idx2] * SNR[idx1:idx2]) / np.sum(SNR[idx1:idx2])
            )
            lF_Hz.append(
                np.sum(F_Hz[idx1:idx2] * SNR[idx1:idx2]) / np.sum(SNR[idx1:idx2])
            )
            lSNR.append(np.sum(SNR[idx1:idx2] ** 2) ** 0.5)
            lSNRpr.append(np.sum(SNR_phase_rel[idx1:idx2] ** 2) ** 0.5)
            if emphasis is not None:
                lemph.append(np.sum(emphasis[idx1:idx2] ** 2) ** 0.5)

        F_Hz = np.asarray(lF_Hz)
        data = np.asarray(ldata)
        SNR_prev = SNR
        SNR = np.asarray(lSNR)
        SNR_phase_rel = np.asarray(lSNRpr)
        SNR = SNR * (np.mean(SNR_prev ** 2) / np.mean(SNR ** 2)) ** 0.5
        if emphasis is not None:
            emphasis = np.asarray(lemph)
    ###################################### downsampling #######################

    inverse_data = grab_kwargs(
        aid, kw, arguments.adjustments.kw_hints, "inverse", kwput=kwput
    )
    if inverse_data:
        data = 1 / data

    F_boost_Hz = grab_kwargs(
        aid, kw, arguments.adjustments.kw_hints, "F_boost_Hz", kwput=kwput
    )
    if F_boost_Hz is not None:
        if emphasis is None:
            emphasis = np.ones_like(SNR)
        for F_s, F_e in F_boost_Hz:
            select_Fb = (F_Hz >= F_s) & (F_Hz < F_e)
            emphasis[select_Fb] *= 2

    grab_kwarg_hints(aid, kw, arguments.adjustments.kw_hints, kwput=kwput)
    grab_kwarg_hints(aid, kw, arguments.ranges.kw_hints, kwput=kwput)

    ZPKrep = representations.ZPKwData(
        F_Hz=F_Hz,
        data=data,
        W=SNR,
        zeros=zeros,
        poles=poles,
        gain=gain,
        zeros_overlay=zeros_overlay,
        poles_overlay=poles_overlay,
        F_nyquist_Hz=F_nyquist_Hz,
        residuals_log_im_scale=SNR_phase_rel,
    )

    if aid.hint("prune_Qrank") is not None:
        aid.fitter_update(
            fitters_ZPK.ZPKrep2MRF(
                ZPKrep,
                residuals_type=aid.hint("residuals_type"),
                coding_map=fitters_ZPK.coding_maps.RI,
                distance_limit_auto=2,
                delay_s=aid.hint("delay_s"),
                delay_s_min=aid.hint("delay_s"),
                delay_s_max=aid.hint("delay_s"),
                h_infinity=aid.hint("h_infinity"),
                h_infinity_deweight=aid.hint("h_infinity_deweight"),
                max_BW_Hz=aid.hint("root_bandwidth_Hz_max"),
                F_cutoff_Hz=aid.hint("root_F_Hz_max"),
            ),
            validate=False,
        )
        with aid.log_heading("prune Q-ranked order reduction"):
            order_reduce.order_reduce(
                aid=aid,
                Q_rank_cutoff=aid.hint("prune_Qrank"),
                reduce_c=True,
                reduce_r=True,
                optimize=False,
            )
            ZPKrep = aid.fitter.ZPKrep

    aid.fitter_update(
        fitters_ZPK.ZPKrep2MRF(
            ZPKrep,
            residuals_type=aid.hint("residuals_type"),
            coding_map=aid.hint("coding_map"),
            distance_limit_auto=2,
            delay_s=aid.hint("delay_s"),
            delay_s_min=aid.hint("delay_s"),
            delay_s_max=aid.hint("delay_s"),
            h_infinity=aid.hint("h_infinity"),
            h_infinity_deweight=aid.hint("h_infinity_deweight"),
            max_BW_Hz=aid.hint("root_bandwidth_Hz_max"),
            F_cutoff_Hz=aid.hint("root_F_Hz_max"),
        ),
        validate=False,
    )

    if kw:
        arguments.check_remaining_arguments(kw, arguments.kw_hints)

    # TODO, put relative degree check in if the USER is specifying a filter and
    # rational fitting will not be employed

    if mode == "dumpargs_full":
        settings = dict(aid.hints)
        settings["zeros"] = zeros
        settings["poles"] = poles
        settings["gain"] = gain
        settings["zeros_overlay"] = zeros_overlay
        settings["poles_overlay"] = poles_overlay
        settings["F_nyquist_Hz"] = F_nyquist_Hz
        return settings
    elif mode == "dumpargs":
        return kwput
    elif mode == "full":
        baseline_order = fit_full(aid, emphasis)
    elif mode == "fullAAA":
        baseline_order = fit_fullAAA(aid, emphasis)
    elif mode == "full2x":
        baseline_order = fit_full2x(aid, emphasis)
    elif mode == "chebydebug":
        with aid.factorization():
            rational_fits.fit_cheby(aid)
        baseline_order = 100
    elif mode == "chebydebug+":
        with aid.factorization():
            rational_fits.fit_cheby(aid)
        order_reduce.order_reduce(
            aid=aid,
            Q_rank_cutoff=0.2,
            optimize=False,
            reduce_c=True,
            reduce_r=True,
        )
        aid.fitter.optimize(aid=aid)
        aid.fitter_update(representative=True)
        baseline_order = 100
    elif mode == "discdebug":
        rational_fits.fit_disc(aid)
        baseline_order = 100
    elif mode == "rational":
        baseline_order = fit_rational(aid, emphasis)
    elif mode == "AAA":
        baseline_order = fit_AAA(aid, emphasis)
    elif mode == "onlyAAA":
        baseline_order = fit_AAAonly(aid, emphasis)
    elif mode == "onlyAAAreduce":
        baseline_order = fit_AAAonly_reduce(aid, emphasis)
    elif mode == "rational2x":
        baseline_order = fit_rational2x(aid, emphasis)
    elif mode == "reduce":
        baseline_order = fit_reduce(aid, emphasis)
    elif mode == "fit":
        baseline_order = fit_only(aid, emphasis)
    elif mode == "copy":
        baseline_order = fit_copy(aid, emphasis)
    elif mode == "gain":
        baseline_order = fit_copy(aid, emphasis)
        with aid.fitter.with_codings_only([aid.fitter.gain_coding]):
            aid.fitter.optimize()
        aid.fitter_update(representative=True)
    else:
        raise RuntimeError("Wrong Option for mode")

    resaid = results_aid_adv.ResultsAidAdv(aid, kw=kwput)
    resaid.choose(baseline_order)
    print("BASELINE: ", baseline_order)
    assert resaid.fitter is not None
    ptbl = resaid.investigate_order_console(print_function=None)
    with aid.log_heading("investigations"):
        aid.log_info(2, ptbl)
    return resaid


def fit_full(aid, emphasis):
    SNR_adjustments.SNR_fix(aid)

    aid.fitter_update(representative=True, validate=True)
    _fit_rational(aid, emphasis)

    return _reduce(aid)


def fit_full2x(aid, emphasis):
    SNR_adjustments.SNR_fix(aid)

    _fit_rational(aid, emphasis, _phase_patch=False, order_hint="order_first")
    _reduce(aid, with_successive=False)
    _fit_rational(aid, emphasis, _phase_patch=True)
    return _reduce(aid)


def _fit_rational(aid, emphasis, _phase_patch=True, order_hint=None):
    if emphasis is None:

        def fit_call():
            rational_fits.fit_cheby(aid, order_hint=order_hint)
            aid.log_progress(
                3,
                "Initial Order: (Z={0}, P={1}, Z-P={2})".format(
                    len(aid.fitter.zeros),
                    len(aid.fitter.poles),
                    aid.fitter.order_relative,
                ),
            )

            assert(np.all(aid.fitter.poles.fullplane.real < 0))
            order_reduce.order_reduce(
                aid=aid,
                Q_rank_cutoff=0.4,
                optimize=False,
            )

            aid.log_progress(
                3,
                "Fastdrop Order: (Z={0}, P={1}, Z-P={2})".format(
                    len(aid.fitter.zeros),
                    len(aid.fitter.poles),
                    aid.fitter.order_relative,
                ),
            )

            phase_patch.root_stabilize(aid)
            if _phase_patch:

                aid.log_progress(4, "mag fitting and phase patching")
                aid.invalidate_fitters()
                phase_patch.phase_patch(aid)
                aid.fitter_update(representative=True)

    else:
        # not sure I like this emphasis mechanism
        def fit_call():
            rational_fits.fit_cheby(aid, order_hint=order_hint)
            aid.log_progress(
                3,
                "Initial Order: (Z={0}, P={1}, Z-P={2})".format(
                    len(aid.fitter.zeros),
                    len(aid.fitter.poles),
                    aid.fitter.order_relative,
                ),
            )

            order_reduce.order_reduce(
                aid=aid,
                Q_rank_cutoff=0.4,
                optimize=False,
            )

            aid.log_progress(
                3,
                "Fastdrop Order: (Z={0}, P={1}, Z-P={2})".format(
                    len(aid.fitter.zeros),
                    len(aid.fitter.poles),
                    aid.fitter.order_relative,
                ),
            )

            with aid.log_heading("Applying emphasis"):
                aid.log_progress(4, "adjusting weights")
                aid.fitter.W = aid.fitter.W * emphasis
                aid.invalidate_fitters()

                aid.fitter.optimize(aid=aid)
                aid.fitter_update(representative=False)

                algorithms.optimize_anneal(aid)
                #aid.fitter_checkup()
                aid.fitter_checkpoint()
                aid.log_progress(4, "rational fitting (more)")
                aid.fitter_update(representative=False)

                # use the previous fits just as a suggestion
                rational_fits.fit_cheby(aid)

                order_reduce.order_reduce(
                    aid=aid,
                    Q_rank_cutoff=0.4,
                    optimize=False,
                )

            if _phase_patch:
                aid.invalidate_fitters()
                phase_patch.root_stabilize(aid)
                aid.log_progress(4, "mag fitting and phase patching")
                phase_patch.phase_patch(aid)
                aid.fitter_update(representative=True)

    with aid.log_heading("rational fitting"):
        if aid.hint("suggest"):
            # use the existing filter only as a suggestion
            fit_call()
        else:
            # apply the rational fit to the factorized form and stabilize only
            # the new cheby fit roots, not the original filter
            with aid.factorization():
                fit_call()

    aid.fitter.optimize(aid=aid)
    aid.fitter_update(representative=True)

    algorithms.optimize_anneal(aid)
    #aid.fitter_checkup()
    aid.fitter_checkpoint()
    return


def fit_fullAAA(aid, emphasis):
    SNR_adjustments.SNR_fix(aid)

    aid.fitter_update(representative=True, validate=True)
    _fit_rational(aid, emphasis, _phase_patch=True)
    _reduce(aid, with_flip=False, with_successive=False, representative=True,)
    aid.fitter_update(
        representative=True,
    )
    _fit_AAA_successive(aid, emphasis, _phase_patch=False)
    return _reduce(aid)


def fit_AAA(aid, emphasis, _phase_patch=False):
    SNR_adjustments.SNR_fix(aid)

    aid.fitter_update(representative=True, validate=True)
    _fit_AAA_successive(aid, emphasis, _phase_patch=_phase_patch)

    with aid.log_heading("Q-ranked order reduction"):
        order_reduce.order_reduce(
            aid=aid,
            Q_rank_cutoff=0.3,
        )
        aid.log_progress(4, "order reduced annealing")
        algorithms.optimize_anneal(aid)
        #aid.fitter_checkup()
        aid.fitter_checkpoint()
    return _reduce(aid)


def fit_AAAonly(aid, emphasis):
    SNR_adjustments.SNR_fix(aid)

    _fit_onlyAAA(aid, emphasis)

    aid.log_progress(
        3,
        "Final Order: (Z={0}, P={1}, Z-P={2}), maxzp {3}".format(
            len(aid.fitter.zeros),
            len(aid.fitter.poles),
            aid.fitter.order_relative,
            aid.fitter_orders().maxzp
        ),
    )
    return aid.fitter_orders().maxzp


def fit_AAAonly_reduce(aid, emphasis):
    SNR_adjustments.SNR_fix(aid)

    _fit_onlyAAA(aid, emphasis)

    return _reduce(aid)


def _fit_onlyAAA(aid, emphasis, order_hint=None):
    def fit_call():

        #do a rational fit first to get reldeg to work
        rational_fits.fit_cheby_base(aid, order=2, order_max=4, order_min=2)
        aid.log_progress(
            3,
            "Initial Order: (Z={0}, P={1}, Z-P={2})".format(
                len(aid.fitter.zeros),
                len(aid.fitter.poles),
                aid.fitter.order_relative,
            ),
        )

        order_reduce.order_reduce(
            aid=aid,
            Q_rank_cutoff=0.4,
            optimize=False,
        )

        with aid.factorization():
            AAA_fits.fit_AAA(aid, order_hint=order_hint)

        aid.log_progress(
            3,
            "Initial Order: (Z={0}, P={1}, Z-P={2})".format(
                len(aid.fitter.zeros),
                len(aid.fitter.poles),
                aid.fitter.order_relative,
            ),
        )

        order_reduce.order_reduce(
            aid=aid,
            Q_rank_cutoff=0.2,
            optimize=False,
        )

        aid.log_progress(
            3,
            "Fastdrop Order: (Z={0}, P={1}, Z-P={2})".format(
                len(aid.fitter.zeros),
                len(aid.fitter.poles),
                aid.fitter.order_relative,
            ),
        )
    with aid.factorization():
        fit_call()
    aid.fitter_update(representative=True, validate=True)

    aid.fitter.optimize(aid=aid)
    aid.fitter_update(representative=True)

    algorithms.optimize_anneal(aid)
    #aid.fitter_checkup()
    aid.fitter_checkpoint(hint_name='AAA')
    aid.fitter_update(representative=True, validate=True)

    aid.log_progress(
        3,
        "Final Order: (Z={0}, P={1}, Z-P={2})".format(
            len(aid.fitter.zeros),
            len(aid.fitter.poles),
            aid.fitter.order_relative,
        ),
    )
    return aid.fitter_orders().maxzp


def _fit_AAA(aid, emphasis, _phase_patch=True, order_hint=None):
    def fit_call():

        rational_fits.fit_cheby_base(aid, order=2, order_max=4, order_min=2)
        aid.log_progress(
            3,
            "Initial Order: (Z={0}, P={1}, Z-P={2})".format(
                len(aid.fitter.zeros),
                len(aid.fitter.poles),
                aid.fitter.order_relative,
            ),
        )

        order_reduce.order_reduce(
            aid=aid,
            Q_rank_cutoff=0.4,
            optimize=False,
            reduce_c=False,
            reduce_r=False,
        )

        aid.log_progress(
            3,
            "Initial OrderB: (Z={0}, P={1}, Z-P={2})".format(
                len(aid.fitter.zeros),
                len(aid.fitter.poles),
                aid.fitter.order_relative,
            ),
        )
        with aid.factorization():
            AAA_fits.fit_AAA(aid, order_hint=order_hint)
        aid.log_progress(
            3,
            "Initial Order: (Z={0}, P={1}, Z-P={2})".format(
                len(aid.fitter.zeros),
                len(aid.fitter.poles),
                aid.fitter.order_relative,
            ),
        )

        order_reduce.order_reduce(
            aid=aid,
            Q_rank_cutoff=0.2,
            optimize=False,
        )

        aid.log_progress(
            3,
            "Fastdrop Order: (Z={0}, P={1}, Z-P={2})".format(
                len(aid.fitter.zeros),
                len(aid.fitter.poles),
                aid.fitter.order_relative,
            ),
        )

        with aid.factorization():
            rational_fits.fit_cheby_base(aid, order=2, order_max=4, order_min=2)

        order_reduce.order_reduce(
            aid=aid,
            Q_rank_cutoff=0.2,
            optimize=False,
        )

        aid.log_progress(
            3,
            "Fastdrop Order: (Z={0}, P={1}, Z-P={2})".format(
                len(aid.fitter.zeros),
                len(aid.fitter.poles),
                aid.fitter.order_relative,
            ),
        )

        #phase_patch.root_stabilize(aid)

        #if _phase_patch:
        #    aid.log_progress(4, "mag fitting and phase patching")
        #    aid.invalidate_fitters()
        #    phase_patch.phase_patch(aid)
        #    aid.fitter_update(representative=True)

    with aid.log_heading("rational fitting"):
        if aid.hint("suggest"):
            # use the existing filter only as a suggestion
            fit_call()
        else:
            # apply the rational fit to the factorized form and stabilize only
            # the new cheby fit roots, not the original filter
            with aid.factorization():
                fit_call()
    aid.fitter_update(representative=True, validate=True)

    aid.fitter.optimize(aid=aid)
    aid.fitter_update(representative=True)

    algorithms.optimize_anneal(aid)
    #aid.fitter_checkup()
    aid.fitter_checkpoint()
    return aid.fitter_orders().maxzp


def _fit_AAA_successive(aid, emphasis, _phase_patch=False, order_hint=None):
    def fit_initial():
        rational_fits.fit_cheby_base(aid, order=2, order_max=4, order_min=2)
        aid.log_progress(
            3,
            "Initial Order: (Z={0}, P={1}, Z-P={2})".format(
                len(aid.fitter.zeros),
                len(aid.fitter.poles),
                aid.fitter.order_relative,
            ),
        )

        order_reduce.order_reduce(
            aid=aid,
            Q_rank_cutoff=0.4,
            optimize=False,
        )
        algorithms.optimize_anneal(aid)
        #aid.fitter_update(representative=True)
        aid.fitter_checkpoint(hint_name='AAA')

    def fit_callAAA():
        resavg1 = aid.fitter.residuals_average
        aid.log_progress("AAA sequence, residuals {0:.2f}".format(resavg1))

        with aid.factorization(data_mod=False):
            AAA_fits.fit_AAA_base(aid, 20, 20, 20)

        order_reduce.order_reduce(
            aid=aid,
            Q_rank_cutoff=0.2,
            optimize=False,
        )

        algorithms.optimize_anneal(aid)
        aid.fitter_checkpoint(hint_name='AAA')

        with aid.factorization(data_mod=False):
            rational_fits.fit_cheby_base(aid, order=2, order_max=2, order_min=2)
        aid.fitter_checkpoint(hint_name='AAA')

        aid.log_progress(
            3,
            "Initial Order: (Z={0}, P={1}, Z-P={2})".format(
                len(aid.fitter.zeros),
                len(aid.fitter.poles),
                aid.fitter.order_relative,
            ),
        )

        order_reduce.order_reduce(
            aid=aid,
            Q_rank_cutoff=0.1,
            optimize=False,
        )

        algorithms.optimize_anneal(aid)
        aid.log_progress(
            3,
            "Fastdrop Order: (Z={0}, P={1}, Z-P={2})".format(
                len(aid.fitter.zeros),
                len(aid.fitter.poles),
                aid.fitter.order_relative,
            ),
        )

        _reduce(aid, with_flip=False, with_successive=False, representative=False, hint_name='AAA')
        # use a checkup and revert before the flip!
        aid.fitter_checkup(hint_name='AAA')
        _reduce(aid, with_flip=False, with_successive=True, representative=False, hint_name='AAA')

        aid.fitter.optimize(aid=aid)
        algorithms.optimize_anneal(aid)
        aid.fitter_checkpoint(hint_name='AAA')

        # this will revert the filter if it is not improved
        improved = aid.fitter_checkup(hint_name='AAA')

        resavg2 = aid.fitter.residuals_average
        aid.log_progress(3, "AAA sequence 2, residuals {0:.2f} from {1:.1f}".format(resavg2, resavg1), )

        if resavg2/resavg1 < 1:
            if aid.fitter_orders().maxzp > aid.hint(
                "rational_AAA_fit_order_max",
                "rational_fit_order_max",
                "order_max",
            ):
                return

            with aid.factorization():
                fit_callAAA()
        return


    with aid.log_heading("rational fitting"):
        if aid.hint("suggest"):
            # use the existing filter only as a suggestion
            fit_initial()
            fit_callAAA()
        else:
            # apply the rational fit to the factorized form and stabilize only
            # the new cheby fit roots, not the original filter
            with aid.factorization():
                fit_initial()
                fit_callAAA()

            aid.fitter_checkup(hint_name='AAA')
            #aid.fitter_checkpoint(hint_name='AAA')

    if _phase_patch:
        assert(False)
        phase_patch.root_stabilize(aid)

        aid.log_progress(4, "mag fitting and phase patching")
        aid.invalidate_fitters()
        phase_patch.phase_patch(aid)
        aid.fitter_update(representative=True)

    aid.log_progress("AAA DONE WITH SEQUENCE".format())
    aid.fitter_checkup(hint_name='AAA')
    aid.fitter_update(representative=True, validate=True)

    aid.fitter.optimize(aid=aid)
    aid.fitter_update(representative=False)

    algorithms.optimize_anneal(aid)
    #aid.fitter_checkup()
    aid.fitter_checkpoint(hint_name='AAA')
    return aid.fitter_orders().maxzp


def fit_rational(aid, emphasis, _phase_patch=True):
    SNR_adjustments.SNR_fix(aid)

    _fit_rational(aid, emphasis, _phase_patch=_phase_patch)

    with aid.log_heading("Q-ranked order reduction"):
        order_reduce.order_reduce(
            aid=aid,
            Q_rank_cutoff=0.7,
        )
        aid.log_progress(4, "order reduced annealing")
        algorithms.optimize_anneal(aid)
        #aid.fitter_checkup()
        aid.fitter_checkpoint()

    return aid.fitter_orders().maxzp


def fit_rational2x(aid, emphasis, _phase_patch=True):
    SNR_adjustments.SNR_fix(aid)

    _fit_rational(aid, emphasis, _phase_patch=False, order_hint="order_first")
    _reduce(aid, with_flip=True, with_successive=False)
    _fit_rational(aid, emphasis, _phase_patch=_phase_patch)

    with aid.log_heading("Q-ranked order reduction"):
        order_reduce.order_reduce(
            aid=aid,
            Q_rank_cutoff=0.7,
        )
        aid.log_progress(4, "order reduced annealing")
        algorithms.optimize_anneal(aid)
        #aid.fitter_checkup()
        aid.fitter_checkpoint()

    return aid.fitter_orders().maxzp


def fit_reduce(aid, emphasis):
    SNR_adjustments.SNR_fix(aid)

    if emphasis is not None:
        aid.fitter.W = aid.fitter.W * emphasis
        aid.invalidate_fitters()
        aid.fitter.optimize(aid=aid)
        aid.fitter_update(representative=True)

        algorithms.optimize_anneal(aid)
        #aid.fitter_checkup()
        aid.fitter_checkpoint()

    with aid.fitter.with_codings_only([aid.fitter.gain_coding]):
        aid.fitter.optimize()
        aid.fitter_update(representative=True)
    aid.fitter.optimize(aid=aid)
    aid.fitter_update(representative=True)

    #aid.fitter_checkup()
    aid.fitter_checkpoint()

    return _reduce(aid)


def fit_only(aid, emphasis):
    SNR_adjustments.SNR_fix(aid)

    if emphasis is not None:
        aid.fitter.W = aid.fitter.W * emphasis
        aid.invalidate_fitters()
        with aid.fitter.with_codings_only([aid.fitter.gain_coding]):
            aid.fitter.optimize()
        aid.fitter_update(representative=True)

        # algorithms.optimize_anneal(aid)
        aid.fitter.optimize(aid=aid)
        aid.fitter_update(representative=True)

        algorithms.optimize_anneal(aid)
        #aid.fitter_checkup()
        aid.fitter_checkpoint()
    else:
        with aid.fitter.with_codings_only([aid.fitter.gain_coding]):
            aid.fitter.optimize()
        aid.fitter_update(representative=True)

    aid.fitter.optimize(aid=aid)
    aid.fitter_update(representative=True)

    algorithms.optimize_anneal(aid)
    #aid.fitter_checkup()
    aid.fitter_checkpoint()

    if aid.hint("delay_s_max") is not None:
        aid.log_progress(4, "activating delay fitting")
        # print('delay_coding: ', aid.fitter.delay_s_max)
        # print('delay_coding: ', aid.fitter.codings_revision)
        aid.fitter.delay_s_max = aid.hint("delay_s_max")
        aid.fitter.delay_s_min = aid.hint("delay_s_min")
        algorithms.optimize_anneal(aid)
        # print('delay_coding: ', aid.fitter.codings_revision)
        # print('delay_coding: ', aid.fitter.delay_s_max)
        # print('delay_coding: ', aid.fitter.delay_coding.N_parameters)
        aid.fitter_checkup()

    algorithms.optimize_anneal(aid)
    #aid.fitter_checkup()
    aid.fitter_checkpoint()

    if aid.hint("delay_s_max") is not None:
        aid.log_alert(2, "Baseline fit delay: ", aid.fitter.delay_s)
    baseline_order = aid.fitter_orders().maxzp
    aid.log_alert(
        2,
        ("Baseline fit residuals: {:.2e}, at order {}").format(
            aid.fitter.residuals_average, baseline_order
        ),
    )
    return baseline_order


def fit_copy(aid, emphasis):
    SNR_adjustments.SNR_fix(aid)
    aid.fitter_update(representative=True)
    baseline_order = aid.fitter_orders().maxzp
    aid.log_alert(
        2,
        ("Baseline fit residuals: {:.2e}, at order {}").format(
            aid.fitter.residuals_average, baseline_order
        ),
    )
    return baseline_order


def _reduce(
    aid,
    with_flip=True,
    with_successive=True,
    representative=True,
    hint_name=None,
):
    baseline_order = aid.fitter_orders().maxzp

    with aid.log_heading("Q-ranked order reduction"):
        rzp_list = order_reduce.order_reduce(
            aid=aid,
            Q_rank_cutoff=2,
        )
        aid.log_progress(4, "order reduced annealing")
        algorithms.optimize_anneal(aid)
        improved = aid.fitter_checkpoint(hint_name=hint_name)

    with aid.log_heading("Q-ranked selective order restoration"):
        if improved:
            rzp_list.sort()
            order_reduce.order_restore(aid, rzp_list)

    algorithms.optimize_anneal(aid)
    #aid.fitter_checkup()
    aid.fitter_checkpoint(hint_name=hint_name)

    if with_flip:
        order_reduce_flip.order_reduce_flip(
            aid=aid,
            non_mindelay=True,
            representative=representative,
            hint_name=hint_name,
        )

    baseline_order = aid.fitter_orders().maxzp
    with aid.log_heading("selective order reduction"):
        order_reduce.order_reduce_selective(
            aid=aid,
            num_total_max=8,
            num_type_max=2,
            representative=representative,
            hint_name=hint_name,
        )

    if aid.hint("delay_s_max") is not None:
        aid.log_progress(4, "activating delay fitting")
        aid.fitter.delay_s_max = aid.hint("delay_s_max")
        aid.fitter.delay_s_min = aid.hint("delay_s_min")
        algorithms.optimize_anneal(aid)
        aid.fitter_checkup(hint_name=hint_name)

    algorithms.optimize_anneal(aid)
    #aid.fitter_checkup(hint_name=hint_name)
    aid.fitter_checkpoint(hint_name=hint_name)

    if aid.hint("delay_s_max") is not None:
        aid.log_alert(2, "Baseline fit delay: ", aid.fitter.delay_s)
    baseline_order = aid.fitter_orders().maxzp

    aid.log_alert(
        2,
        ("Baseline fit residuals: {:.2e}, at order {}").format(
            aid.fitter.residuals_average, baseline_order
        ),
    )

    if aid.hint("baseline_only"):
        return baseline_order
    # TODO, now need to make a choice based on residuals
    if with_successive:
        with aid.log_heading("successive order reduction"):
            order_reduce.order_reduce_successive(
                aid=aid,
                num_total_max=6,
                num_type_max=2,
                representative=representative,
                hint_name=hint_name,
            )
    return baseline_order


# Sphinx complains a lot about this module documentation
if "sphinx" not in sys.modules:
    data2filter.__doc__ = autogen_docstr.__doc__



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import collections
from wield import declarative
import numpy as np

from ... import fitters_ZPK
from ...utilities import args
from ... import representations


class ArgumentError(ValueError):
    pass


def mapcheck_bool(aid, aname, val):
    if isinstance(val, str):
        if val.lower() in ["true", "yes", "1"]:
            val = True
        elif val.lower() in ["false", "no", "0"]:
            val = False
        else:
            raise ArgumentError(
                ("Argument {} has unrecognized bool specifier {}").format(aname, val)
            )
    return bool(val)


def float_check(aid, aname, val):
    try:
        return float(val)
    except ValueError:
        raise ArgumentError(("argument {}={} must be a float").format(aname, val))


def mapcheck_float(aid, aname, val):
    try:
        return float(val)
    except ValueError:
        raise ArgumentError(("argument {}={} must be a float").format(aname, val))


def mapcheck_positive_float(aid, aname, val):
    val = float(val)
    if not (val > 0):
        raise ArgumentError(("argument {}={} must be positive").format(aname, val))
    return val


def mapcheck_nonnegative_float(aid, aname, val):
    val = float(val)
    if not (val >= 0):
        raise ArgumentError(("argument {}={} must not be negative").format(aname, val))
    return val


def mapcheck_nonnegative_float_orNone(aid, aname, val):
    if isinstance(val, str):
        if val.lower() in ["none", "null"]:
            val = None
    if val is None:
        return val
    return mapcheck_nonnegative_float(aid, aname, val)


def mapcheck_positive_float_orNone(aid, aname, val):
    if isinstance(val, str):
        if val.lower() in ["none", "null"]:
            val = None
    if val is None:
        return val
    return mapcheck_positive_float(aid, aname, val)


def mapcheck_int(aid, aname, val):
    try:
        val = int(val)
    except ValueError:
        raise ArgumentError(("argument {}={} must be an integer").format(aname, val))

    return val


def mapcheck_positive_int(aid, aname, val):
    try:
        val = int(val)
    except ValueError:
        raise ArgumentError(("argument {}={} must be an integer").format(aname, val))

    # TODO error reporting
    if not (val > 0):
        raise ArgumentError(("argument {}={} must be positive").format(aname, val))
    if not (val - float(val) == 0):
        raise ArgumentError(("argument {}={} must be an integer").format(aname, val))
    return val


def mapcheck_nonnegative_int(aid, aname, val):
    try:
        val = int(val)
    except ValueError:
        raise ArgumentError(("argument {}={} must be an integer").format(aname, val))

    # TODO error reporting
    if not (val >= 0):
        raise ArgumentError(("argument {}={} must be non-negative").format(aname, val))
    if not (val - float(val) == 0):
        raise ArgumentError(("argument {}={} must be an integer").format(aname, val))
    return val


def mapcheck_positive_int_orNone(aid, aname, val):
    if isinstance(val, str):
        if val.lower() in ["none", "null"]:
            val = None
    if val is None:
        return val
    return mapcheck_positive_int(aid, aname, val)


def mapcheck_nonnegative_int_orNone(aid, aname, val):
    if isinstance(val, str):
        if val.lower() in ["none", "null"]:
            val = None
    if val is None:
        return val
    return mapcheck_nonnegative_int(aid, aname, val)


def mapcheck_int_orNone(aid, aname, val):
    if isinstance(val, str):
        if val.lower() in ["none", "null"]:
            val = None
    if val is None:
        return val
    return mapcheck_int(aid, aname, val)


def kw_ZPKrep_build(args, kwargs):
    """ """
    kw = dict()
    ZPKrep = None
    for arg in args:
        if isinstance(arg, collections.Mapping):
            kw.update(arg)
        elif isinstance(
            arg,
            (
                representations.ZPKwData,
                fitters_ZPK.MultiReprFilterBase,
            ),
        ):
            if ZPKrep is None:
                ZPKrep = arg.ZPKrep
            else:
                raise ArgumentError(
                    "Only a single ZPKrep may be specified" " in positional arguments"
                )
        else:
            raise ArgumentError(
                "Positional Arguments are not allowed,"
                " except for a single master ZPKrep, or a set of"
                " default keyword argument dictionaries"
            )
    kw.update(kwargs)
    return kw, ZPKrep


def grab_kwarg_hints(aid, kw, kwdesc, kwput=None):
    def eval_hint(hint_name):
        kwmeta = kwdesc[hint_name]

        reqs = kwmeta.get("require_hints", None)
        if reqs is not None:
            for hname in reqs:
                if not aid.hint_has(hname):
                    eval_hint(hname)
                if not aid.hint_has(hname):
                    raise RuntimeError("Hint/Kwarg evaluation out of order. BUG!")

        found = _grab_kwargs(aid, kw, kwmeta, hint_name, kwput=kwput)
        if found:
            aid.hint_setdefault(hint_name, next(iter(found.values())))
        # TODO, make better error message for case when default is not given
        try:
            default = kwmeta["default"]
        except KeyError:
            pass
        else:
            if callable(default):
                default = default(aid, hint_name)
            aid.hint_setdefault(hint_name, default)

    for hint_name in kwdesc.keys():
        eval_hint(hint_name)


def grab_kwargs(aid, kw, kwdesc, argname, kwput=None):
    kwmeta = kwdesc[argname]
    found = _grab_kwargs(aid, kw, kwmeta, argname, kwput=kwput)
    if found:
        return next(iter(found.values()))

    # TODO, make better error message for case when default is not given
    default = kwmeta["default"]
    if callable(default):
        default = default(aid, argname)
    return default


def _grab_kwargs(aid, kw, kwmeta, argname, kwput=None):
    reqs = kwmeta.get("require_hints", None)
    if reqs is not None:
        for hname in reqs:
            if not aid.hint_has(hname):
                raise RuntimeError("Hint/Kwarg evaluation out of order. BUG!")
    drop_val = kwmeta.get("drop_val", args.UNSPEC)
    drop_vals = list(kwmeta.get("drop_vals", []))
    if drop_val is not args.UNSPEC:
        drop_vals.extend(drop_val)

    pop = kwmeta.get("pop", True)
    mapcheck = kwmeta.get("mapcheck", None)
    found = {}
    # use a list to modify it from the function
    discrepancy = [False]

    def check_find(aname):
        if pop:
            val = kw.pop(aname, args.UNSPEC)
        else:
            val = kw.get(aname, args.UNSPEC)

        if val is args.UNSPEC:
            return False

        if kwput is not None:
            kwput[aname] = val

        if val in drop_vals:
            return False

        if mapcheck is not None:
            val = mapcheck(aid, aname, val)

        if found:
            if np.any(next(iter(found.values())) != val):
                discrepancy[0] = True
        found[aname] = val

        return True

    prefname = kwmeta.get("name", argname)
    check_find(prefname)

    for aname in kwmeta.get("aliases", []):
        check_find(aname)

    for aname in kwmeta.get("aliases_bad", []):
        if check_find(aname):
            aid.log_warn(
                2,
                "Argument '{}' is a deprecated alias."
                " Use '{}' instead".format(aname, prefname),
            )

    if discrepancy[0]:
        kv_strs = []
        for k, v in found.items():
            kv_strs.append("{}: {}".format(k, v))
        raise ArgumentError(
            "Inconsistent values given for aliased arguments\n" + "\n".join(kv_strs)
        )

    return found


def check_remaining_arguments(kw, kwdict):
    names = set()
    alias_map = {}
    bad_map = {}
    for hname, hdict in kwdict.items():
        name = hdict.get("name", hname)
        names.add(name)
        for alias in hdict.get("aliases", []):
            alias_map[alias] = name
        for alias in hdict.get("aliases_bad", []):
            bad_map[alias] = name

    allnames = set(names)
    allnames.update(alias_map.keys())
    allnames.update(bad_map.keys())
    import difflib

    k_lines = dict()
    for k in kw.keys():
        l = []
        k_lines[k] = l
        matches = difflib.get_close_matches(k, allnames, n=5)
        print("Matches: ", k, matches)
        mainset = set()
        for match in matches:
            if match in names:
                mainset.add(name)
                l.append("'{}'".format(match))
            elif match in alias_map:
                mainmatch = alias_map[match]
                l.append("'{}', alias for '{}'".format(match, mainmatch))
            else:
                mainmatch = bad_map[match]
                l.append("'{}', deprecated alias for '{}'".format(match, mainmatch))
    stack_lines = []
    for k, lines in sorted(k_lines.items()):
        if len(lines) > 0:
            stack_lines.append("'{}':\n\t".format(k) + "\n\t".join(lines))
        else:
            stack_lines.append("'{}': <no similar arguments recognized>".format(k))
    raise ArgumentError(
        (
            "Unrecognized keyword arguments. Listed with potential matches below,"
            "\nor call with argument 'help'=True for more details\n"
            "{}"
        ).format("\n".join(stack_lines))
    )
    return


def cplx_iIjJ(val):
    val = val.replace("i", "j")
    val = val.replace("I", "j")
    val = val.replace("J", "j")
    return complex(val)


def cplx_iIjJ_list_inner(val):
    vals = val.split()
    vals2 = []
    for v in vals:
        vals2.extend(v.split(","))
    vals3 = []
    for v in vals2:
        vals3.extend(v.split(";"))
    return [cplx_iIjJ(v) for v in vals3 if v != ""]


def cplx_iIjJ_list(val):
    vals = val.split(":")
    val = vals[0]
    vals = vals[1:]
    val = cplx_iIjJ_list_inner(val)
    if vals:
        val_c = vals[0]
        vals = vals[1:]
        assert(not vals)
        val_c = cplx_iIjJ_list_inner(val_c)
        val.extend(val_c)
        val.extend(np.asarray(val_c).conjugate())
    return val


def transfer_kw(kwA, kwB, kwdict, pop=False):
    for hname, hdict in kwdict.items():
        if pop:
            name = hdict.pop("name", hname)
        else:
            name = hdict.get("name", hname)
        normalize = hdict.get("normalize", None)
        val = kwA.get(name, declarative.NOARG)
        if val is not declarative.NOARG:
            if normalize is not None:
                val = normalize(val)
            kwB[name] = val

        for aname in hdict.get("aliases", []):
            if pop:
                val = kwA.pop(aname, declarative.NOARG)
            else:
                val = kwA.get(aname, declarative.NOARG)
            if val is not declarative.NOARG:
                if normalize is not None:
                    val = normalize(val)
                kwB[aname] = val

        for aname in hdict.get("aliases_bad", []):
            if pop:
                val = kwA.pop(aname, declarative.NOARG)
            else:
                val = kwA.get(aname, declarative.NOARG)
            if val is not declarative.NOARG:
                if normalize is not None:
                    val = normalize(val)
                kwB[aname] = val



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


from . import codings_s
from . import codings_z

from .codings_z import BranchCutAmbiguity

from .MRF import (
    MultiReprFilterBase,
    MultiReprFilterZ,
    MultiReprFilterS,
)


from .ZPKrep2MRF import (
    ZPKrep2MRF,
    MRF2MRF,
)

from .mappings import coding_maps



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

try:
    from collections.abc import Mapping as MappingABC
except ImportError:
    from collections import Mapping as MappingABC


def argscan(*args, **kwargs):
    """
    The attibute name must be given as keyword argument "arg". Giving the name
    allows it to be looked-up in the argument dictionary or locals().
    """
    argname = kwargs.get("arg", None)
    for val in args:
        if val is REQ:
            raise TypeError(
                "Argument {0} is required and must be supplied to above function".format(
                    argname
                )
            )
        elif (argname is not None) and isinstance(val, MappingABC):
            m_arg = val.get(argname, UNSPEC)
            if m_arg is not UNSPEC:
                return m_arg
        elif val is not UNSPEC:
            return val
    # return the last value, likely None or the loop would have returned it
    return val


# Special unique value to specify that arguments are required or not yet specified
REQ = (argscan, "REQUIRED")
UNSPEC = (argscan, "UNSPECIFIED")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

# import warnings
import contextlib
from wield import declarative

from .. import fitters_ZPK
from .. import annotate
from .. import plots
from . import algorithms


def fitter_order(fitter):
    return max(len(fitter.zeros), len(fitter.poles))


class FitterUnacceptable(Exception):
    pass


class FitAid(object):
    def __init__(
        self,
        doc_db,
        hints=None,
        residuals_type="log",
    ):
        self.fitter = None
        self.fitter_current = None
        self.fitter_lowres_avg = None
        self.fitter_lowres_max = None
        self.fitter_lowres_med = None
        self.fitter_loword = None

        self.doc_db_method = None

        if hints is None:
            hints = dict()

        if isinstance(hints, (list, tuple)):
            usehints = dict()
            for hdict in hints:
                usehints.update(hdict)
        else:
            usehints = dict(hints)

        self.hints = usehints
        self.hints_defaults = dict()

        self.doc_db = doc_db
        self.doc_stack = []

        self.section = [0]
        self.section_stack = []

        # TODO make this a hint
        self.verbosity = 5

        return

    @property
    def residuals_type(self):
        return self.hint("residuals_type", default="log")

    def _check_residuals(self, fitter):
        if not isinstance(fitter, fitters_ZPK.MultiReprFilterBase):
            raise FitterUnacceptable()
        if fitter.residuals_type == self.residuals_type:
            if fitter.residuals_log_im_scale != 1:
                f_new = fitter.copy(
                    residuals_type=self.residuals_type,
                    residuals_log_im_scale=1,
                )
                return f_new
        return None

    def hint_set(self, hname, hval, default=False):
        if default:
            self.hints.setdefault(hname, hval)
        else:
            self.hints[hname] = hval
        return

    def hint_arg(self, func_arg, *args, **kwargs):
        if func_arg is None:
            return self.hint(*args, **kwargs)
        else:
            return func_arg

    def hint(self, *args, **kwargs):
        superarg = []
        for arg in args:
            if isinstance(arg, (list, tuple)):
                superarg.extend(arg)
            else:
                superarg.append(arg)

        default = kwargs.get("default", declarative.NOARG)
        if default is declarative.NOARG:
            raise RuntimeError("Must have default return")

        first_hint = superarg[0]

        lh_set = self.hints_defaults.get(first_hint, None)
        if lh_set is None:
            lh_set = set()
            self.hints_defaults[first_hint] = lh_set
        lh_set.add(default)

        for key in superarg:
            key = key.format(**kwargs)
            ret = self.hints.setdefault(key, None)
            if ret is not None:
                return ret
        return default

    def fitter_check(
        self,
        fitter_new=None,
        hint_name=None,
        update=True,
        variant=None,
        validate=True,
        annotate="unknown",
    ):
        """ """

        if variant is None:
            raise RuntimeError(
                'Must specify variant for fitter_check, such as ["OrdUp", "OrdDn", "OrdC"]'
            )

        if self.fitter is None:
            if update:
                self.fitter = fitter_new
                self.fitter_update()
            else:
                return True

        if fitter_new is None:
            fitter_new = self.fitter

        # TODO, should check that it is using standard residuals

        resavgN = lambda: fitter_new.residuals_average
        resmaxN = lambda: fitter_new.residuals_max
        resmedN = lambda: fitter_new.residuals_med
        resavgC = lambda: self.fitter_current.residuals_average
        resmaxC = lambda: self.fitter_current.residuals_max
        resmedC = lambda: self.fitter_current.residuals_med
        resavgB = lambda: self.fitter_lowres_avg.residuals_average
        resmaxB = lambda: self.fitter_lowres_max.residuals_max
        resmedB = lambda: self.fitter_lowres_med.residuals_med

        print("resavgC:", resavgC())

        if hint_name is None:
            hint_name = "default"

        improved = [True]

        def check_hint(
            hints,
            calc,
        ):
            hintval = self.hint(hints, default=None)
            if hintval is not None and calc > hintval:
                self.log("Failed {}, {} > {}".format(hints[0], calc, hintval))
                improved[0] = False

        check_hint(
            [
                "resavg_Ethresh{0}.{1}".format(variant, hint_name),
                "resavg_Ethresh{0}".format(variant),
            ],
            resavgN(),
        )

        check_hint(
            [
                "resmax_Ethresh{0}.{1}".format(variant, hint_name),
                "resmax_Ethresh{0}".format(variant),
            ],
            resmaxN(),
        )

        check_hint(
            [
                "resavg_Rthresh{0}.{1}".format(variant, hint_name),
                "resavg_Rthresh{0}".format(variant),
            ],
            (resavgN() / resavgC()),
        )

        check_hint(
            [
                "resmax_Rthresh{0}.{1}".format(variant, hint_name),
                "resmax_Rthresh{0}".format(variant),
            ],
            (resmaxN() / resmaxC()),
        )

        check_hint(
            [
                "resmed_Rthresh{0}.{1}".format(variant, hint_name),
                "resmed_Rthresh{0}".format(variant),
            ],
            (resmedN() / resmedC()),
        )

        check_hint(
            [
                "resavgB_Rthresh{0}.{1}".format(variant, hint_name),
                "resavgB_Rthresh{0}".format(variant),
            ],
            (resavgN() / resavgB()),
        )

        check_hint(
            [
                "resmaxB_Rthresh{0}.{1}".format(variant, hint_name),
                "resmaxB_Rthresh{0}".format(variant),
            ],
            (resmaxN() / resmaxB()),
        )

        check_hint(
            [
                "resmedB_Rthresh{0}.{1}".format(variant, hint_name),
                "resmedB_Rthresh{0}".format(variant),
            ],
            (resmedN() / resmedB()),
        )

        self.log("IMPROVED: ", improved[0])
        if improved[0] and update:
            self.fitter_update(
                fitter_new,
                annotate=None,
            )
            if annotate is not None:
                self.annotate(
                    name="fitter_check (improved)",
                    about=annotate,
                    fitter=fitter_new,
                    plotter=plots.plot_fitter_flag,
                    method=self.doc_db_method,
                    verbosity=6,
                )
        else:
            if annotate is not None:
                self.annotate(
                    name="fitter_check (not improved)",
                    fitter=fitter_new,
                    about=annotate,
                    plotter=plots.plot_fitter_flag,
                    method=self.doc_db_method,
                    verbosity=10,
                )
        val_func = self.hint("fitter_check_validate", default=None)
        if validate and val_func is not None and fitter_new is not None:
            val_func(self, fitter_new)
        return improved[0]

    def fitter_checkup(self):
        improved = self.fitter_check(
            self.fitter,
            variant="OrdC",
        )
        if not improved:
            self.fitter = self.fitter_current.copy()
        return improved

    def fitter_update(
        self,
        fitter=None,
        validate=True,
        annotate="unknown",
    ):
        if fitter is None:
            fitter = self.fitter
        else:
            self.fitter = fitter
        avg1 = fitter.residuals_average
        algorithms.sign_check_flip(fitter)
        avg2 = fitter.residuals_average
        assert avg2 <= avg1

        if annotate is not None:
            self.annotate(
                name="fitter_update",
                about=annotate,
                fitter=fitter,
                plotter=plots.plot_fitter_flag,
                method=self.doc_db_method,
                verbosity=8,
            )

        val_func = self.hint("fitter_update_validate", default=None)
        if validate and val_func is not None and fitter is not None:
            val_func(self, fitter)

        try:
            fitter_use = self._check_residuals(fitter)
        except FitterUnacceptable:
            return

        if fitter_use is not None:
            fitter_use.optimize(aid=self)
            fitter_copy = [fitter_use]
        else:
            fitter_use = fitter
            fitter_copy = []

        def copy1():
            if not fitter_copy:
                fitter_copy.append(fitter.copy())
            return fitter_copy[0]

        # print("ZPK", fitter_use.ZPKsf)

        new_res = fitter_use.residuals_average
        new_res_max = fitter_use.residuals_max
        new_res_med = fitter_use.residuals_med
        new_ord = fitter_order(fitter_use)

        if self.fitter_current is not None:
            old_res = self.fitter_lowres_avg.residuals_average
            old_res_max = self.fitter_lowres_max.residuals_max
            old_res_med = self.fitter_lowres_med.residuals_med
            old_ord = fitter_order(self.fitter_loword)

            if new_res < old_res:
                self.fitter_lowres_avg = copy1()

            if new_res_max < old_res_max:
                self.fitter_lowres_max = copy1()

            if new_res_med < old_res_med:
                self.fitter_lowres_med = copy1()

            if new_ord < old_ord:
                self.fitter_loword = copy1()

            self.fitter_current = copy1()
        else:
            self.fitter_current = copy1()
            self.fitter_lowres_avg = copy1()
            self.fitter_lowres_max = copy1()
            self.fitter_lowres_med = copy1()
            self.fitter_loword = copy1()

        # print("UPDATED: ", self.fitter_current.residuals_average)

        return fitter_use

    def log(self, *args, **kwargs):
        # hint = kwargs.get('hint', None)

        if self.hint("no_log", default=False):
            return

        if self.hint("log_print", default=True):
            print(*args)

        # TODO put in the doc_db

    def digest_write(self, folder, format="markdown", **kwargs):
        if format in ["md", "markdown"]:
            return annotate.docdb2markdown(folder, self.doc_db, **kwargs)
        else:
            raise RuntimeError(
                "Format Not Recognized, must be one of: ['md', 'markdown']"
            )

    @contextlib.contextmanager
    def annotate_into(self, name):
        doc_db_prev = self.doc_db

        self.doc_db = annotate.annotate_into(
            self.doc_db,
            name=name,
        )
        yield

        self.doc_db = doc_db_prev
        return

    @contextlib.contextmanager
    def annotate_method(self, name):
        doc_db_method_prev = self.doc_db_method

        self.doc_db_method = name
        yield

        self.doc_db_method = doc_db_method_prev
        return

    def annotate(
        self,
        name,
        verbosity,
        about,
        method=None,
        verbosity_limit=None,
        fitter=None,
        plotter=None,
    ):
        if method is None:
            method = self.doc_db_method
        annotate.annotate(
            self.doc_db,
            name=name,
            verbosity=verbosity,
            method=method,
            about=about,
            verbosity_limit=verbosity_limit,
            fitter=fitter,
            plotter=plotter,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


# import numpy as np

from ..codings_cmn import (
    CodingType,
    # Ipi,
    # I2pi
)

# import scipy.linalg


class CodingFBW(CodingType):
    N_parameters = 2
    p_BW_Hz = 0
    p_F_Hz = 0

    def update(self, real, imag):
        self.p_BW_Hz = real
        self.p_F_Hz = imag

    def reduce(self):
        return [
            self.p_BW_Hz,
            self.p_F_Hz,
        ]

    def update_roots(self, r1):
        """
        r2, may be unspecified, in which case it is assumed to be nothing, if r1 is real, or otherwise the conjugate of r1
        """
        self.p_BW_Hz = -r1.real
        self.p_F_Hz = r1.imag
        return

    @property
    def gain_effect(self):
        if self.p_BW_Hz > 0:
            return 1
        else:
            return 1

    def transfer(self):
        # real/imaginary part of root
        r, i = -self.p_BW_Hz, self.p_F_Hz
        X = self.sys.Xsf_grid
        Xsq = self.sys.Xsf_grid_sq
        return (r * r + i * i) - 2 * X * r + Xsq

    def derivative(self):
        if self.disable:
            return []
        # real/imaginary part of root
        r, i = -self.p_BW_Hz, self.p_F_Hz
        X = self.sys.Xsf_grid
        return [
            -2 * (r - X),
            (2 * i),
        ]

    def roots_c(self):
        # real/imaginary part of root
        r, i = -self.p_BW_Hz, abs(self.p_F_Hz)
        return [r + 1j * i]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import numpy as np
import collections
from wield.bunch import Bunch

from .base import (
    ArgumentError,
    mapcheck_bool,
    mapcheck_float,
    mapcheck_positive_float_orNone,
    cplx_iIjJ_list,
)

from ... import representations


def check_F_Hz(aid, hname, val):
    val = np.asarray(val)
    if np.any(val < 0):
        raise ArgumentError("Only positive frequencies allowed in 'F_Hz'")


def list_flatten(l):
    l_flat = []

    def flatten(l):
        if isinstance(l, (list, tuple)):
            for subl in l:
                flatten(subl)
        elif isinstance(l, np.ndarray):
            l_flat.extend(l)
        else:
            l_flat.append(l)

    flatten(l)
    return l_flat


def normalize_roots(val):
    if isinstance(val, str):
        if val.lower() == "none":
            val = None
        else:
            val = cplx_iIjJ_list(val)
    elif isinstance(val, collections.Mapping):
        r = val["real"]
        i = val["imag"]
        val = np.asarray(r) + 1j * np.asarray(i)
    if val is None:
        return None
    # TODO, treat the single number case better
    val = list_flatten([val])
    return val


def mapcheck_roots(aid, hname, val):
    try:
        val = normalize_roots(val)
        complete = aid.hint("pair_completion")
        return representations.asMRRB(val, complete=complete)
    except ValueError:
        raise ArgumentError(
            ("Argument {} must be a list of complex numbers").format(hname, val)
        )


def mapcheck_ZPK(aid, hname, val):
    if val is None:
        return None
    complete = aid.hint("pair_completion")
    return representations.asZPKTF(val, complete=complete)


def mapcheck_mode(aid, hname, val):
    modes = [
        "full",
        "fullAAA",
        "full2x",
        "fit",
        "AAA",
        "onlyAAA",
        "onlyAAAreduce",
        "reduce",
        "rational",
        "rational2x",
        "dumpargs",
        "chebydebug",
        "chebydebug+",
        "discdebug",
        "copy",
        "gain",
    ]
    if val not in modes:
        raise ArgumentError("Argument {} must be one of {}".format(hname, modes))
    return val


kw_hints_pre = Bunch(
    pair_completion=dict(
        mapcheck=mapcheck_bool,
        default=True,
    )
)


kw_hints_data = Bunch(
    data=dict(
        APgroup="fitting",
        # TODO, mapcheck
        aliases=["xfer"],
        aliases_bad=["TF", "d", "transfer"],
        about="""
        Data of the transfer function to fit. Should be complex. Can only fit
        transfer functions on real data (conjugate negative frequencies).
        """,
    ),
    F_Hz=dict(
        APgroup="fitting",
        # TODO, mapcheck
        aliases=["f_Hz", "frequency_Hz"],
        aliases_bad=["freq", "F", "f", "frequency"],
        about="""
        Frequencies for the transfer function data. Should only be positive.
        """,
        checkmap=check_F_Hz,
    ),
    SNR=dict(
        APgroup="fitting",
        # TODO, mapcheck
        aliases=["W", "snr"],
        aliases_bad=["weight", "weights"],
        default=None,
        about="""
        Statistical weights used for fitting and decision procedures.
        Values not trusted and rescaled adjusted in default algorithms. Use
        the 'emphasis' parameter to adjust weighting between statistical
        decisions and accuracy/implementation requirements for the fit.
        """,
    ),
    coherence=dict(
        APgroup="fitting",
        # TODO, mapcheck
        aliases=["coherence", "coh", "COH"],
        default=None,
        about="""
        Coherence of transfer functions used. This is converted into SNR.
        """,
    ),
    SNR_phase_relative=dict(
        APgroup="fitting",
        # TODO, mapcheck
        aliases=["Wphase_relative", "SNR_phase_rel"],
        default=None,
        about="""
        Statistical weights specifically for the phase part of the fit. This
        only affects the ZPK fitting (second half) of "full" mode, and the
        "fit", "reduce" and "gain" modes. The default of "None" specifies to
        use the SNR for both types of fit. This is a relative scaling to adjust
        the SNR by for the phase aspect of the fit.
        """,
    ),
    emphasis=dict(
        # TODO, mapcheck
        APgroup="fitting",
        aliases=["Wemph"],
        default=None,
        about="""
        Emphasis weighting. Affects fits and some decision procedures differently
        than the purely statistical data. Not rescaled or adjusted by algorithms.
        Use values with small dynamic range as the reduction in the effective
        number of data points scales with the dynamic range of the emphasis squared.
        """,
    ),
    select=dict(
        APignore=True,
        # TODO, mapcheck
        default=None,
        about="""
        This may be a slice object or boolean array to remove points from: data,
        F_Hz, SNR, and emphasis. May be anything that can index numpy arrays.
        """,
    ),
)

kw_hints_fit = Bunch(
    help=dict(
        APignore=True,
        aliases=["h"],
        mapcheck=mapcheck_bool,
        default=False,
        about="""
        Print a help screen
        """,
    ),
    # TODO, added new modes to docs
    mode=dict(
        APgroup="fitting",
        APpriority=3,
        mapcheck=mapcheck_mode,
        default="full",
        about="""
        Fitting mode to use, to change the automation level. Must be one of
         - "full": to use rational fitting to get initial parameters, then alternate
                   optimization and order reduction
         - "rational": to use rational fitting to get initial parameters and then
                       stop after the simplest order reduction. Useful to use with
                       other fitters. Does not perform delay fitting.
         - "fit": To take an initial ZPK guess, and only fit/optimize it. Good
                  for refining previous fits.
         - "reduce": To take an initial ZPK guess, and then alternate fitting and
                     order reduction.
         - "dumpargs": Return dictionary of arguments to see the full run settings.
        """,
    ),
    chain=dict(
        APgroup="fitting",
        APignore=True,
        APpriority=3,
        # mapcheck = mapcheck_chain,
        default=None,
        about="""
        Chain argument to run sequential fits. Takes a fitter results object.
        This will change the default fitting from mode='full' to
        mode='fit_reduce', but will keep mode='fit'. This jumpstarts all
        defaults to use those supplied for the previous fit. In addition, the
        ZPKs of the previous fit are used. Any additional arguments will override
        the defaults. This can be used to modify the fitting SNR/weights, or
        possibly to activate H_infinity weighting.
        """,
    ),
    F_nyquist_Hz=dict(
        APshort="-N",
        APignore=False,
        # TODO, mapcheck
        APpriority=5,
        mapcheck=mapcheck_positive_float_orNone,
        default=None,
        about="""
        This selects the Nyquist frequency for Z-domain fitting. If None or not
        specified, the fits are done in the "Sf" domain - the s domain scaled for
        units of frequency.
        """,
    ),
    ZPK=dict(
        APignore=True,
        mapcheck=mapcheck_ZPK,
        require_hints=["pair_completion"],
        aliases=["zpk"],
        default=None,
    ),
    zeros=dict(
        APshort="-z",
        APnargs="*",
        APgroup="fitting",
        APtype=cplx_iIjJ_list,
        mapcheck=mapcheck_roots,
        normalize=normalize_roots,
        require_hints=["pair_completion"],
        aliases=["Z", "z"],
        default=None,
        about="Initial zeros used in the fit.",
    ),
    poles=dict(
        APshort="-p",
        APgroup="fitting",
        APnargs="*",
        APtype=cplx_iIjJ_list,
        mapcheck=mapcheck_roots,
        normalize=normalize_roots,
        require_hints=["pair_completion"],
        aliases=["P", "p"],
        default=None,
        about="Initial poles used in the fit.",
    ),
    gain=dict(
        APshort="-k",
        APgroup="fitting",
        APtype=float,
        mapcheck=mapcheck_float,
        aliases=["K", "k"],
        default=None,
        about="Initial gain for the fit",
    ),
    ZPK_overlay=dict(
        APignore=True,
        mapcheck=mapcheck_ZPK,
        require_hints=["pair_completion"],
        aliases=["zpk_overlay"],
        default=None,
    ),
    zeros_overlay=dict(
        APshort="-Z",
        APgroup="fitting",
        APnargs="*",
        APtype=cplx_iIjJ_list,
        mapcheck=mapcheck_roots,
        normalize=normalize_roots,
        require_hints=["pair_completion"],
        aliases=["Zov"],
        default=None,
        about="zeros guaranteed to be part of the fit.",
    ),
    poles_overlay=dict(
        APshort="-P",
        APgroup="fitting",
        APnargs="*",
        APtype=cplx_iIjJ_list,
        mapcheck=mapcheck_roots,
        normalize=normalize_roots,
        require_hints=["pair_completion"],
        aliases=["Pov"],
        default=None,
        about="poles guaranteed to be part of the fit.",
    ),
)

kw_hints = dict()
kw_hints.update(kw_hints_fit)
kw_hints.update(kw_hints_data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


from ..codings_cmn import (
    CodingType,
    # Ipi,
    # I2pi
)


class CodingSOS(CodingType):
    N_parameters = 2
    p_c1 = 0
    p_c2 = 0
    single_root = False

    def update(self, c1, c2=None):
        if self.single_root:
            self.p_c1 = c1
        else:
            self.p_c1 = c1
            self.p_c2 = c2

    def reduce(self):
        if self.single_root:
            return [self.p_c1]
        else:
            return [self.p_c1, self.p_c2]

    def transfer(self):
        if self.single_root:
            # second order sections (2x roots either real or complex conj)
            X = self.sys.Xsf_grid
            return self.p_c1 + X
        else:
            # second order sections (2x roots either real or complex conj)
            X = self.sys.Xsf_grid
            Xsq = self.sys.Xsf_grid_sq
            return self.p_c2 + X * self.p_c1 + Xsq

    def derivative(self):
        if self.single_root:
            return [1]
        else:
            X = self.sys.Xsf_grid
            # Xsq = self.sys.Xsf_grid_sq
            return [X, 1]

    @property
    def gain_effect(self):
        rs = self.roots_r()
        if len(rs) == 0:
            return 1
        elif len(rs) == 1:
            return 1 if rs[0] < 0 else -1
        else:
            if rs[0] < 0:
                return 1 if rs[1] < 0 else -1
            else:
                return 1 if rs[1] > 0 else -1

    def update_roots(self, r1, r2=None):
        """
        r2, may be unspecified, in which case it is assumed to be nothing, if r1 is real, or otherwise the conjugate of r1
        """
        if r2 is None and r1.imag == 0:
            self.single_root = True
            self.N_parameters = 1
            self.p_c1 = -r1.real
        else:
            self.single_root = False
            self.N_parameters = 2
            if r1.imag != 0:
                # TODO check conjugates
                self.p_c2 = r1.real ** 2 + r1.imag ** 2
                self.p_c1 = -2 * r1.real
            else:
                self.p_c2 = r1 * r2
                self.p_c1 = -(r1 + r2)
        return

    def roots(self):
        # second order sections (2x roots either real or complex conj)
        if self.single_root:
            return [-self.p_c1]
        else:
            # a = self.p_c2, b = self.p_c1, c = 1
            disc = self.p_c1 * self.p_c1 - 4 * self.p_c2
            if disc >= 0:
                sqrt_disc = disc ** 0.5
                if self.p_c1 < 0:
                    r1 = (-self.p_c1 + sqrt_disc) / 2
                else:
                    r1 = (-self.p_c1 - sqrt_disc) / 2
                r2 = self.p_c2 / r1
                return [r1, r2]
            else:
                sqrt_disc = (-disc) ** 0.5
                r1 = (-self.p_c1 + sqrt_disc * 1j) / 2
                return [r1, r1.conjugate()]

    def roots_r(self):
        # second order sections (2x roots either real or complex conj)
        if self.single_root:
            return [-self.p_c1]
        else:
            # a = self.p_c2, b = self.p_c1, c = 1
            disc = self.p_c1 * self.p_c1 - 4 * self.p_c2
            if disc >= 0:
                sqrt_disc = disc ** 0.5
                if self.p_c1 < 0:
                    r1 = (-self.p_c1 + sqrt_disc) / 2
                else:
                    r1 = (-self.p_c1 - sqrt_disc) / 2
                r2 = self.p_c2 / r1
                return [r1, r2]
            else:
                return []

    def roots_c(self):
        # second order sections (2x roots either real or complex conj)
        if self.single_root:
            return []
        else:
            # complex form is always numerically stable
            # a = self.p_c2, b = self.p_c1, c = 1
            disc = self.p_c1 * self.p_c1 - 4 * self.p_c2
            if disc >= 0:
                sqrt_disc = disc ** 0.5
                return []
            else:
                sqrt_disc = (-disc) ** 0.5
                r1 = (-self.p_c1 + sqrt_disc * 1j) / 2
                return [r1]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

from scipy._lib._util import _asarray_validated
from scipy.linalg.misc import LinAlgError, _datacopied
from scipy.linalg.blas import find_best_blas_type
from scipy.linalg.lapack import _compute_lwork

from . import lapack_svd


def svd_full(
    a,
    full_matrices=True,
    compute_u=True,
    compute_v=True,
    overwrite_a=False,
    check_finite=True,
):
    """ """
    a1 = _asarray_validated(a, check_finite=check_finite)
    if len(a1.shape) != 2:
        raise ValueError("expected matrix")
    m, n = a1.shape
    overwrite_a = overwrite_a or (_datacopied(a1, a))

    prefix, dtype, prefer_fortran = find_best_blas_type(arrays=(a,))
    if prefix == "d":
        gesXd, gesXd_lwork = lapack_svd.dgesvd, lapack_svd.dgesvd_lwork
    elif prefix == "s":
        gesXd, gesXd_lwork = lapack_svd.sgesvd, lapack_svd.sgesvd_lwork
    elif prefix == "c":
        gesXd, gesXd_lwork = lapack_svd.cgesvd, lapack_svd.cgesvd_lwork
    else:
        gesXd, gesXd_lwork = lapack_svd.zgesvd, lapack_svd.zgesvd_lwork

    # compute optimal lwork
    lwork = _compute_lwork(
        gesXd_lwork,
        a1,
        compute_u=compute_u,
        compute_v=compute_v,
        full_matrices=full_matrices,
    )

    # perform decomposition
    u, s, v, info = gesXd(
        a1,
        compute_u=compute_u,
        compute_v=compute_v,
        lwork=lwork,
        full_matrices=full_matrices,
        overwrite_a=overwrite_a,
    )

    if info > 0:
        raise LinAlgError("SVD did not converge")
    if info < 0:
        raise ValueError("illegal value in %d-th argument of internal gesdd" % -info)
    if compute_u and compute_v:
        return u, s, v
    elif compute_u:
        return u, s
    elif compute_v:
        return s, v
    else:
        return s


def svd(
    a,
    compute_u=True,
    compute_v=True,
    il=None,
    iu=None,
    overwrite_a=False,
    check_finite=True,
):
    """ """
    # check if the bounds check is being used
    if il is None and iu is None:
        return svd_full(
            a=a,
            compute_u=compute_u,
            compute_v=compute_v,
            overwrite_a=overwrite_a,
            check_finite=check_finite,
        )
    minmn = min(a.shape[0], a.shape[1])

    if il is None:
        il = 1
    elif il < 0:
        if il < -minmn:
            il = -minmn
        il = minmn + il + 1
    else:
        # change to fortran convention
        il += 1
        if il > minmn:
            il = minmn

    if iu is None:
        iu = minmn
    elif iu < 0:
        iu = minmn + iu + 1
    else:
        # change to fortran convention
        iu += 1
        if iu > minmn:
            iu = minmn

    a1 = _asarray_validated(a, check_finite=check_finite)
    if len(a1.shape) != 2:
        raise ValueError("expected matrix")
    m, n = a1.shape
    overwrite_a = overwrite_a or (_datacopied(a1, a))

    prefix, dtype, prefer_fortran = find_best_blas_type(arrays=(a,))
    if prefix == "d":
        gesXd, gesXd_lwork = lapack_svd.dgesvdx, lapack_svd.dgesvdx_lwork
    elif prefix == "s":
        gesXd, gesXd_lwork = lapack_svd.sgesvdx, lapack_svd.sgesvdx_lwork
    elif prefix == "c":
        gesXd, gesXd_lwork = lapack_svd.cgesvdx, lapack_svd.cgesvdx_lwork
    else:
        gesXd, gesXd_lwork = lapack_svd.zgesvdx, lapack_svd.zgesvdx_lwork

    # compute optimal lwork
    lwork = _compute_lwork(
        gesXd_lwork,
        a1,
        compute_u=compute_u,
        compute_v=compute_v,
        il=il,
        iu=iu,
    )

    # perform decomposition
    u, ns, s, v, iwork, info = gesXd(
        a1,
        compute_u=compute_u,
        compute_v=compute_v,
        il=il,
        iu=iu,
        lwork=lwork,
        overwrite_a=overwrite_a,
    )

    if info > 0:
        raise LinAlgError("SVD did not converge")
    if info < 0:
        raise ValueError("illegal value in %d-th argument of internal gesdd" % -info)
    if compute_u and compute_v:
        return u[:ns], s[:ns], v[:ns]
    elif compute_u:
        return u[:ns], s[:ns]
    elif compute_v:
        return s[:ns], v[:ns]
    else:
        return s[:ns]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import matplotlib.pyplot as plt
from wield import declarative

from . import plot_fit
from . import plot_residuals
from . import overlays
from . import ZP
from . import flags
from . import rel_comparison


class IIRPlots(declarative.OverridableObject):
    # this magic causes the OverridableObject constructor arguments
    # to be SAVED into _overridable_object_save_kwargs
    # this allows convenient copy/update methods
    _overridable_object_save_kwargs = True
    _overridable_object_kwargs = None

    zeroID_color = "darkorange"
    zeroOD_color = "darkgreen"
    poleID_color = "blue"
    poleOD_color = "red"
    label_data = "data"
    label_fit = "fit ({nP}P, {nZ}Z, Rsq ={Rsq:.2e})"

    show_immediate = False
    show_saved = False
    force_save = False

    def new(self, fname=None, **kwargs):
        kw = dict(self._overridable_object_kwargs)
        kw.update(kwargs)
        return self.__class__(**kw)

    def show(self):
        plt.show()

    def _post_plot(self, axB, fname, kwargs):
        do_show = True

        # don't show if in a subcommand (where fig is already specified to render within)
        if (
            kwargs.get("ax_bunch", None) is not None
            or kwargs.get("fig", None) is not None
        ):
            do_show = False

        if fname is not None:
            axB.save(fname)
            if not self.show_saved:
                plt.close(axB.fig)
            elif do_show and self.show_immediate:
                plt.show()
        elif self.force_save:
            raise RuntimeError("Filename to save not specified and force_save active")
        elif do_show and self.show_immediate:
            plt.show()

    def plot_fit(self, fitter, fname=None, **kwargs):
        axB = plot_fit.plot_fit(self, fitter, **kwargs)
        self._post_plot(axB, fname, kwargs)
        return axB

    def plot_residuals(self, fitter=None, fname=None, **kwargs):
        axB = plot_residuals.plot_residuals(self, fitter, **kwargs)
        self._post_plot(axB, fname, kwargs)
        return axB

    def plot_rel_comparison(self, fitter, fitter_ref, fname=None, **kwargs):
        axB = rel_comparison.plot_rel_comparison(self, fitter, fitter_ref, **kwargs)
        self._post_plot(axB, fname, kwargs)
        return axB

    def plot_ZP(self, fitter, fname=None, **kwargs):
        axB = ZP.plot_ZP(self, fitter, **kwargs)
        self._post_plot(axB, fname, kwargs)
        return axB

    def plot_ZP_S(self, fitter, fname=None, **kwargs):
        axB = ZP.plot_ZP_S(self, fitter, **kwargs)
        self._post_plot(axB, fname, kwargs)
        return axB

    def plot_fitter_flag(self, fitter, fname=None, **kwargs):
        axB = flags.plot_fitter_flag(self, fitter, **kwargs)
        self._post_plot(axB, fname, kwargs)
        return axB

    def plot_fitter_flag_residuals(self, fitter=None, fname=None, **kwargs):
        axB = flags.plot_fitter_flag_residuals(self, fitter=fitter, **kwargs)
        self._post_plot(axB, fname, kwargs)
        return axB

    def plot_fitter_flag_compare(self, fitter, fitter_ref, fname=None, **kwargs):
        axB = flags.plot_fitter_flag_compare(self, fitter, fitter_ref, **kwargs)
        self._post_plot(axB, fname, kwargs)
        return axB

    def plot_ZP_grab(self, fitter, duals, fname=None, **kwargs):
        axB = overlays.plot_ZP_grab(self, fitter, duals, **kwargs)
        self._post_plot(axB, fname, kwargs)
        return axB



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import pickle


def load_pickle(fname):
    with open(fname) as F:
        fdict = pickle.load(F)
    return fdict


def write_pickle(fname, fdict):
    with open(fname, "wb") as F:
        fdict = pickle.dump(fdict, F)
    return fdict



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

# import numpy as np

# oldprint = print
# def print(*args):
#    pargs = []
#    for a in args:
#        try:
#            if isinstance(a, np.array):
#                rep = a
#            else:
#                rep = unicode(a, 'utf-8')
#        except TypeError:
#            rep = repr(a)
#        pargs.append(rep)
#    oldprint(*args)

try:
    from IPython.lib.pretty import pprint
except ImportError:
    from pprint import pprint



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""
from wield.bunch import Bunch

from ... import fitters_ZPK
from .base import (
    ArgumentError,
    mapcheck_bool,
    mapcheck_positive_float,
    mapcheck_nonnegative_float,
    mapcheck_nonnegative_float_orNone,
    mapcheck_positive_int_orNone,
    mapcheck_nonnegative_int_orNone,
    mapcheck_positive_float_orNone,
    mapcheck_positive_int,
)


def mapcheck_downsample_type(aid, aname, val):
    val = str(val).lower()
    allowed = ["linear", "lin", "log", "linlog", "loglin"]
    if val not in allowed:
        raise ArgumentError("Argument '{}' must be one of {}".format(aname, allowed))
    return val


def mapcheck_residuals_type(aid, aname, val):
    val = str(val)
    allowed = ["log", "dualA", "dualB", "zeros", "poles"]
    if val not in allowed:
        raise ArgumentError("Argument '{}' must be one of {}".format(aname, allowed))
    return val


def default_residuals_type(aid, aname):
    if not aid.hint("alternate_residuals"):
        return "log"
    else:
        return "dualA"


def default_residuals_type_alt(aid, aname):
    if not aid.hint("alternate_residuals"):
        return "dualA"
    else:
        return "log"


def mapcheck_boostlist(aid, aname, val):
    if val is None:
        return None
    try:
        if isinstance(val, str):
            val_split = val.split(",")
            val = []
            for v in val_split:
                a, b = v.split("-")
                a = float(a)
                b = float(b)

            val.append((float(a), float(b)))
    except Exception:
        raise ArgumentError("{} must be a range tuple or list of range tuples")

    if not isinstance(val, (list, tuple)):
        raise ArgumentError("{} must be a range tuple or list of range tuples")

    for v in val:
        if not isinstance(v, (list, tuple)):
            break
        if len(v) != 2:
            raise ArgumentError("{} must be a range tuple or list of range tuples")
    else:
        return val

    if len(val) != 2:
        raise ArgumentError("{} must be a range tuple or list of range tuples")

    return (val,)


kw_hints = Bunch(
    order_max=dict(
        APgroup="order",
        mapcheck=mapcheck_positive_int,
        default=100,
        about="""
        Maximium order that the rational fitter order estimation may use.
        Increasing this value from the default may impact numerical stability.
        """,
    ),
    order_min=dict(
        APgroup="order",
        mapcheck=mapcheck_positive_int,
        default=20,
        about="""
        Minimum order to use during order estimation. Smaller values will speed
        up fits, but may fail to fix complex data.
        """,
    ),
    order_initial=dict(
        APgroup="order",
        mapcheck=mapcheck_positive_int_orNone,
        default=None,
        aliases=[
            "order",
            "rational_fit_order",
            "rational_cheby_fit_order",
        ],
        about="""
        Order to use in rational fitting. If not specified, the order is
        increased until the residuals plateau.
        """,
    ),
    order_first=dict(
        APgroup="order",
        mapcheck=mapcheck_positive_int_orNone,
        default=8,
        aliases=[
            "order1",
        ],
        about="""
        Order to use in rational fitting for the first round of full2x. Can be lower to be fast.
        """,
    ),
    # emphasis_order_max = dict(
    #    APgroup = 'emphasis',
    #    mapcheck = mapcheck_positive_int_orNone,
    #    default = None,
    #    about = """
    #    Maximium order that the rational fitter order estimation may use during
    #    the emphasis application stage.
    #    """,
    # ),
    # emphasis_order_min = dict(
    #    APgroup = 'emphasis',
    #    #TODO, check minmax with emphasis_order_max
    #    mapcheck = mapcheck_positive_int_orNone,
    #    default = None,
    #    about = """
    #    Minimum order to use during order estimation during the emphasis application
    #    stage.
    #    """,
    # ),
    # emphasis_order_initial = dict(
    #    APgroup = 'emphasis',
    #    mapcheck = mapcheck_positive_int_orNone,
    #    default = None,
    #    aliases = [
    #        'emphasis_order',
    #        'emphasis_rational_fit_order',
    #        'emphasis_rational_cheby_fit_order',
    #    ],
    #    about = """
    #    Order to use in rational fitting during the emphasis application stage.
    #    """,
    # ),
    F_boost_Hz=dict(
        APgroup="SNR adjustments",
        mapcheck=mapcheck_boostlist,
        default=None,
        about="""
        start-end frequency pairs (e.g. 1.2-1.6) to double the SNR end to add emphasis during fitting. Multiple can be specified, comma separated.
        """,
    ),
    downsample=dict(
        APgroup="SNR adjustments",
        mapcheck=mapcheck_positive_int_orNone,
        default=None,
        about="""
        Downsample the data to this number of points. Uses SNR weighing to average during downsampling.
        """,
    ),
    downsample_type=dict(
        APgroup="SNR adjustments",
        default="log",
        about="""
        Use this type of spacing for the downsampled points: linear, log, or loglinear.
        """,
    ),
    F_max_Hz=dict(
        APgroup="SNR adjustments",
        mapcheck=mapcheck_positive_float_orNone,
        default=None,
        about="""
        Maximum frequency to use, cuts off data above this frequency.
        """,
    ),
    F_min_Hz=dict(
        APgroup="SNR adjustments",
        mapcheck=mapcheck_positive_float_orNone,
        default=None,
        about="""
        Minimum frequency to use, cuts off data below this frequency.
        """,
    ),
    SNR_max=dict(
        APgroup="SNR Adjustments",
        mapcheck=mapcheck_positive_float_orNone,
        default=None,
        aliases=[
            "W_max",
        ],
        about="""
        Maximum SNR. Hard cutoff for SNR above this level.
        """,
    ),
    SNR_min=dict(
        APgroup="SNR Adjustments",
        mapcheck=mapcheck_nonnegative_float_orNone,
        default=None,
        aliases=[
            "W_min",
        ],
        about="""
        Minimum SNR. Hard cutoff for SNR below this level.
        """,
    ),
    SNR_regularize_ratio=dict(
        APgroup="SNR Adjustments",
        mapcheck=mapcheck_positive_float,
        default=0.5,
        about="""
        Ratio of effective data points, determined by the dynamic range of the
        SNR weighting, to the actual number of data points. This parameter
        adjusts how the SNR is adjusted to ensure sufficient data points for a
        good fit.
        """,
    ),
    SNR_regularize_scale=dict(
        APgroup="SNR Adjustments",
        mapcheck=mapcheck_positive_float,
        default=10,
        about="""
        Similar to SNR_regularize_ratio, but instead of a direct ratio, the ratio
        is determined as 'SNR_regularize_scale' / max(SNR). For very high SNR
        fits, the SNR should be whitened significantly using the ratio adjustment,
        since the error is dominated by systematics rather than statistical noise.
        The default is 10, causing a %90 coverage ratio when the highest SNR is
        100 (1% statistical error).
        """,
    ),
    SNR_estimate_width=dict(
        APgroup="SNR Adjustments",
        mapcheck=mapcheck_nonnegative_int_orNone,
        default=10,
        about="""
        The window width of nearby points to use for averaging and median filtering
        when estimating the SNR via windowed sample variance. Default is 10.
        None or 0 indicates to not try.
        """,
    ),
    trust_SNR=dict(
        APgroup="SNR Adjustments",
        APaction="store_true",
        mapcheck=mapcheck_bool,
        default=False,
        about="""
        Overall parameter determining if SNR/W statistical weighting input should
        be trusted. If False (the default) then additional statistical tests are
        run to adjust the SNR for better fits. If the user can fully trust their
        SNR estimators and fits to be unbiased, then the user likely does not
        need this tool.
        """,
    ),
    inverse=dict(
        APgroup="fit",
        APaction="store_true",
        mapcheck=mapcheck_bool,
        default=False,
        about="""
        take the reciprocal/inverse of the data (1/xfer) and fit that. Good for fitting compensation filters.
        """,
    ),
    never_unstable_poles=dict(
        APgroup="fit",
        APaction="store_true",
        mapcheck=mapcheck_bool,
        default=False,
        about="""
        During the phase patching tests, unstable roots will be added if detected
        with statistical significance. This prevents the addition of unstable poles.
        """,
    ),
    never_unstable_zeros=dict(
        APgroup="fit",
        APaction="store_true",
        mapcheck=mapcheck_bool,
        default=False,
        about="""
        During the phase patching tests, unstable roots will be added if detected
        with statistical significance. This prevents the addition of unstable zeros.
        """,
    ),
    suggest=dict(
        APgroup="operating mode",
        APaction="store_true",
        mapcheck=mapcheck_bool,
        default=False,
        about="""
        How to use the rational fitter when provided an initial ZPK argument.
        the default of False causes the rational to overlay the ZPK suggestion,
        forcing the roots in the suggestion to be used during the optimization
        stage. If set to True, the ZPK will suggest initial poles to use during
        the fit, which can accellerate convergence or cause it to start at a
        lower initial order, speeding up future fitting operations.
        """,
    ),
    coding_map=dict(
        APignore=True,
        # TODO, add a mapcheck
        default=fitters_ZPK.codings_s.coding_maps.SOSnl,
        advanced=True,
        about="""
        The coding map for the nonlinear fitter. This determines the nonlinear
        parameterization best suited for the NLS fine-grained optimization. These
        parameterizations can also influence the bounds and tradoffs of roots
        placement.
        """,
    ),
    resavg_RthreshOrdDn=dict(
        APgroup="advanced",
        default=1.10,
        mapcheck=mapcheck_positive_float,
        about="""
        The threshold relative change in the average residuals to accept a fit
        of lower order. Only used for the "baseline" fit determination before
        delay is activated and not used during the total order reduction.
        """,
    ),
    resavg_RthreshOrdUp=dict(
        APgroup="advanced",
        default=0.95,
        mapcheck=mapcheck_positive_float,
        about="""
        The threshold relative change in the average residuals to accept a fit
        of higher order. Only used for the "baseline" fit determination before
        delay is activated and not used during the total order reduction.
        """,
    ),
    resavg_RthreshOrdC=dict(
        APgroup="advanced",
        default=1.00,
        mapcheck=mapcheck_positive_float,
        about="""
        The threshold relative change in the average residuals to accept a fit
        of equal order. Only used for the "baseline" fit determination before
        delay is activated and not used during the total order reduction.
        """,
    ),
    distance_limit_scale=dict(
        APgroup="tunings",
        default=1,
        mapcheck=mapcheck_positive_float,
        about="""
        Scaling for how to limit root bandwidth for roots in-between data points.
        Increase to force roots to lower Q, decrease to allow higher Qs.
        """,
    ),
    alternate_residuals=dict(
        APgroup="residuals",
        default=False,
        mapcheck=mapcheck_bool,
        about="""
        Flag to switch from using log/phase residuals to using dual
        ratio residuals as the principle fit residuals. Log/phase is typically
        better behaved.
        """,
    ),
    h_infinity=dict(
        APgroup="residuals",
        # TODO, may a checker for this
        # mapcheck = mapcheck_nonnegative_float,
        default=0,
        about="""
        Use an partial h_infinity norm during the nonlinear fitting
        (does not apply to the linear fits). This argument takes a [0, 1] float
        for the fraction of the residuals to include. If 1, this is true h_infinity,
        if < 1, then the residuals are ranked, and the smallest fraction of them
        have h_infinity_deweight applied to them, to lower their affect of the fit.
        This argument may also be an array, in which case it reweights the
        ranked residuals (ranking array should be largest to smallest). It may
        alternatively take a function, which is given the length, and returns
        the reweighting array.
        """,
    ),
    h_infinity_deweight=dict(
        APgroup="residuals",
        mapcheck=mapcheck_nonnegative_float,
        default=0.1,
        about="""
        Factor to de-weight the residuals below the h_infinity threshold. If 0,
        they are clipped. Defaults to 0.1, to reduce the significance 10x.
        Factors >1 imply an h_0 type fit, where outliers are more strongly ignored.
        """,
    ),
    residuals_type=dict(
        APgroup="residuals",
        APchoices=["log", "dualA", "dualB", "poles", "zeros"],
        default=default_residuals_type,
        mapcheck=mapcheck_residuals_type,
        advanced=True,
        require_hints=["alternate_residuals"],
        about="""
        Standard residuals type to use for the optimizations. Must be one of the
        definitions below, where R=fit/data, the ratio of fit to data and W is
        the SNR:
            log:   W*(ln(|R|) + 1j*R.imag/|R|)
            dualA: W*(R + 1/R - 2)/2
            dualB: W*(R - 1/R)/2
            poles: W*(1/R - 1)
            zeros: W*(R - 1)
        """,
    ),
    residuals_type_alt=dict(
        APgroup="residuals",
        default=default_residuals_type_alt,
        APchoices=["log", "dualA", "dualB", "poles", "zeros"],
        mapcheck=mapcheck_residuals_type,
        advanced=True,
        require_hints=["alternate_residuals"],
        about="""
        Standard alternate residuals type to use during optimization annealing.
        See residuals_type for options.
        """,
    ),
    prune_Qrank=dict(
        APgroup="tuning",
        default=None,
        mapcheck=mapcheck_positive_float_orNone,
        aliases=[
            "prune",
        ],
        about="""
        Pre-prune the filter. Useful for order reduction of StateSpace models.
        The Q-ranking is like a weighted fminreal order reduction. It ranks
        pole-zero by their distance divided by bandwidth. Values smaller than
        .1-.5 remove only pairs that have small adjustments to a transfer function
        and can generally be removed without modifying it.
        """,
    ),
    root_bandwidth_Hz_max=dict(
        APgroup="tunings",
        default=None,
        mapcheck=mapcheck_positive_float,
        aliases=[
            "max_BW_Hz",
            "BW_max_Hz",
        ],
        aliases_bad=[
            "root_bandwidth_max",
            "BW_max",
            "max_BW",
        ],
        about="""
        Maximum bandwidth of any pole or zero in the fit. This affects the
        asymptotic rolloff when relative_degree is constrained.
        """,
    ),
    root_F_Hz_max=dict(
        APgroup="tunings",
        default=None,
        mapcheck=mapcheck_positive_float,
        aliases=[
            "max_F_Hz",
        ],
        aliases_bad=[
            "root_F_max",
            "F_max",
            "max_F",
        ],
        about="""
        Maximum frequency of root complex part.
        """,
    ),
    greedy_order=dict(
        APgroup="speed",
        default=30,
        mapcheck=mapcheck_positive_int_orNone,
        about="""
        Does only partial optimization during the reduce step. Greatly speeds
        up order reduction, but may not be as effective. Defaults to 30, where
        it uses greedy optimization until reaching order 30, then uses combinatoric
        optimization.
        """,
    ),
    baseline_only=dict(
        APgroup="speed",
        default=False,
        APaction="store_true",
        mapcheck=mapcheck_bool,
        about="""
        Only fit down to baseline, do not do successive order reduction.
        """,
    ),
    multithreading=dict(
        APgroup="computing",
        default=None,
        mapcheck=mapcheck_positive_int_orNone,
        about="""
        Perform the order reduction trials using a thread pool of this size.
        If None (the default) or 1, do not activate the pool. Note that many python/numpy
        implementations use a BLAS library that already multithread linear algebra operations.
        """,
    ),
)

#additional hints that use a form that can't be directly assigned into dict kw.
kw_hints['resavg_RthreshOrdDn.AAA'] = dict(
    APignore=True,
    APgroup="advanced",
    default=1,
    mapcheck=mapcheck_positive_float,
    about="""
        The threshold relative change in the average residuals to accept a fit
        of lower order. This is used during the AAA successive order increase.
        """,
)
kw_hints['resavg_RthreshOrdUp.AAA'] = dict(
    APignore=True,
    APgroup="advanced",
    default=1,
    mapcheck=mapcheck_positive_float,
    about="""
        The threshold relative change in the average residuals to accept a fit
        of higher order. This is used during the AAA successive order increase.
        """,
)
kw_hints['resavg_RthreshOrdC.AAA'] = dict(
    APignore=True,
    APgroup="advanced",
    default=1.00,
    mapcheck=mapcheck_positive_float,
    about="""
        The threshold relative change in the average residuals to accept a fit
        of equal order. Used during the AAA successive fitting
        """,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""
from wield.bunch import Bunch

from .base import (
    ArgumentError,
    mapcheck_positive_float_orNone,
    mapcheck_positive_float,
    mapcheck_int_orNone,
    mapcheck_positive_int_orNone,
)


def mapcheck_reldeg(aid, aname, val):
    val = mapcheck_int_orNone(aid, aname, val)
    if val is None:
        return val
    ival = int(val)
    if ival != float(val):
        raise ArgumentError("argument '{}' must be an integer".format(aname))
    if aname == "relative_degree_min":
        rdmax = aid.hint("relative_degree_max")
        if rdmax is not None and val > rdmax:
            raise ArgumentError(
                "argument 'relative_degree_min'={} must be smaller"
                " or equal to 'relative_degree_max'={}".format(val, rdmax)
            )
    return val


def default_reldeg_minmax(aid, aname):
    val = aid.hint("relative_degree")
    return val


def default_delay_s(aid, aname):
    val_min = aid.hint("delay_s_min")
    val_max = aid.hint("delay_s_max")
    if val_max is not None:
        if val_min < 0:
            if val_max > 0:
                val = 0
            else:
                val = val_max
        else:
            val = val_min
    else:
        val = val_min
    return val


def mapcheck_delay_s(aid, aname, val):
    val = float(val)
    val_min = aid.hint("delay_s_min")
    val_max = aid.hint("delay_s_max")
    if val < val_min:
        raise ArgumentError(
            ("Argument 'delay_s'={} must be larger than 'delay_s_min'={}").format(
                val, val_min
            )
        )

    if val_max is not None:
        if val > val_max:
            raise ArgumentError(
                ("Argument 'delay_s'={} must be larger than 'delay_s_max'={}").format(
                    val, val_max
                )
            )
    elif val > val_min:
        raise ArgumentError(
            (
                "Argument 'delay_s' must be equal to 'delay_s_min' since"
                " 'delay_s_max' is not specified. Either specify 'delay_s_max'"
                " or don't specify 'delay_s', in which case it will default to 'delay_s_min'."
            ).format(val, val_min)
        )
    return val


def mapcheck_total_degree_min(aid, aname, val):
    val = mapcheck_positive_int_orNone(aid, aname, val)
    if val is None:
        return val
    val = int(val)
    if val < 0:
        raise ArgumentError("Argument '{}'=val must not be negative".format(aname, val))
    return val


kw_hints = Bunch(
    relative_degree=dict(
        APgroup="order",
        APpriority=10,
        mapcheck=mapcheck_reldeg,
        default=None,
        about="""
        Sets the initial relative degree (number zeros minus polse).
        Defaults to None, which will be the midpoint of the min and max if they
        are specified, otherwise this will default to 0, which typically will
        still fit filters to the correct degree. If constrained to be different
        than the data, the filter will enter the asymptotic regime set by the
        relative degree within 2x of the 'root_bandwidth_Hz_max' setting.
        """,
    ),
    relative_degree_max=dict(
        APgroup="order",
        APpriority=11,
        mapcheck=mapcheck_reldeg,
        default=default_reldeg_minmax,
        require_hints=["relative_degree"],
        about="""
        Maximum value for the filter relative degree (number zeros minus polse).
        Defaults to the relative degree (which may be None). If this value is
        None, the degree is unconstrained and the fit will land at some degree that fits well.
        """,
    ),
    relative_degree_min=dict(
        APgroup="order",
        APpriority=12,
        mapcheck=mapcheck_reldeg,
        default=default_reldeg_minmax,
        require_hints=[
            "relative_degree",
            # needed for the ordering check
            "relative_degree_max",
        ],
        about="""
        Minimum value for the filter relative degree (number zeros minus polse).
        Defaults to the relative degree (which may be None). If this value is
        None, the degree is unconstrained and the fit will land at some degree that fits well.
        """,
    ),
    total_degree_min=dict(
        APgroup="order",
        APpriority=13,
        mapcheck=mapcheck_total_degree_min,
        default=2,
        aliases=[
            "degree_min",
        ],
        about="""
        Minimum degree to search through during the successive order reduction phase.
        Defaults to 2. If None, then successive reduction will not be performed.
        """,
    ),
    delay_s=dict(
        APgroup="delay",
        APpriority=16,
        mapcheck=mapcheck_delay_s,
        default=default_delay_s,
        aliases=["delay_seconds"],
        aliases_bad=["delay"],
        require_hints=[
            "delay_s_min",
            "delay_s_max",
        ],
        about="""
        Use this delay (in seconds) for the initial fitting up until the
        "baseline" fit determination is complete. By default this is the same
        as delay_s_min, unless delay_s_min is negative, in which case it defaults
        to the smaller of 0 seconds or delay_s_max.
        """,
    ),
    delay_s_max=dict(
        APgroup="delay",
        APpriority=15,
        mapcheck=mapcheck_positive_float_orNone,
        default=None,
        aliases=["delay_max_s"],
        aliases_bad=["delay_max"],
        about="""
        The maximum delay in seconds. Defaults to None, in which case delay is
        never fit as a free parameter.
        """,
    ),
    delay_s_min=dict(
        APgroup="delay",
        APpriority=17,
        mapcheck=mapcheck_positive_float,
        default=0,
        aliases=["delay_min_s"],
        aliases_bad=["delay_min"],
        about="""
        The minimum delay in seconds. Defaults to 0. Also sets the typical
        default value of the delay used for the initial part of the fits.
        Must always be specified (cannot be None).
        """,
    ),
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np
from wield import declarative

from ..fitters_rational import RationalDiscFilter, ChebychevFilter
from .. import annotate
from .. import plots
from ..utilities import args, ensure_aid
from .. import TFmath
from .. import representations

from . import nyquist_move


def rational_disc_fit(
    argB=args.UNSPEC,
    F_Hz=args.UNSPEC,
    data=args.UNSPEC,
    nyquist_final_Hz=args.UNSPEC,
    SNR=args.UNSPEC,
    SNR_cutoff=args.UNSPEC,
    ZPK_overlay=args.UNSPEC,
    order=args.UNSPEC,
    order_max=args.UNSPEC,
    complete=args.UNSPEC,
    doc_db=None,
    aid=None,
):
    aid = ensure_aid(aid)

    F_Hz = args.argscan(locals(), argB, args.REQ, arg="F_Hz")
    data = args.argscan(locals(), argB, args.REQ, arg="data")
    nyquist_final_Hz = args.argscan(locals(), argB, None, arg="nyquist_final_Hz")
    SNR = args.argscan(locals(), argB, args.REQ, arg="SNR")
    SNR_cutoff = args.argscan(locals(), argB, 100, arg="SNR_cutoff")
    ZPK_overlay = args.argscan(locals(), argB, None, arg="ZPK_overlay")
    order = args.argscan(locals(), argB, None, arg="order")
    order_max = args.argscan(locals(), argB, 100, arg="order_max")
    complete = args.argscan(locals(), argB, True, arg="complete")

    if nyquist_final_Hz is not None and nyquist_final_Hz <= 0:
        raise RuntimeError(
            "nyquist_final_Hz must be positive for Z domain or None" " for Sf domain."
        )

    if order is not None:
        order = int(order)

    if order_max is not None:
        order_max = int(order_max)

    N_first = 20
    N_final = int(min(len(F_Hz) // 10, order_max))

    if order is not None:
        return ratdisc_single(
            F_Hz=F_Hz,
            data=data,
            nyquist_final_Hz=nyquist_final_Hz,
            SNR=SNR,
            SNR_cutoff=SNR_cutoff,
            ZPK_overlay=ZPK_overlay,
            order=order,
            doc_db=doc_db,
            aid=aid,
        )

    if doc_db is None:
        doc_db_last = None
        doc_db_current = None

    doc_db_true = doc_db
    # otherwise, scan through
    N_current = N_first

    if doc_db is not None:
        doc_db_last = declarative.DeepBunch()
        doc_db_last.update_recursive(doc_db)
    fitter_last = ratdisc_single(
        F_Hz=F_Hz,
        data=data,
        nyquist_final_Hz=nyquist_final_Hz,
        SNR=SNR,
        SNR_cutoff=SNR_cutoff,
        ZPK_overlay=ZPK_overlay,
        order=N_current,
        complete=complete,
        doc_db=doc_db_last,
        aid=aid,
        no_nyquist_shift=True,
    )
    restot_last = fitter_last.residuals_average

    def fitter_ord(fitter):
        return max(len(fitter.zeros), len(fitter.poles))

    while True:
        if N_current == N_final:
            fitter_use = fitter_last
            break

        N_current = N_current * 2
        if N_current > N_final:
            N_current = N_final

        if doc_db is not None:
            doc_db_current = declarative.DeepBunch()
            doc_db_current.update_recursive(doc_db)
        fitter = ratdisc_single(
            F_Hz=F_Hz,
            data=data,
            nyquist_final_Hz=nyquist_final_Hz,
            SNR=SNR,
            SNR_cutoff=SNR_cutoff,
            ZPK_overlay=ZPK_overlay,
            order=N_current,
            doc_db=doc_db_current,
            complete=complete,
            aid=aid,
            no_nyquist_shift=True,
        )
        restot = fitter.residuals_average
        fitter_red = fitter.copy()
        fitter_red.matched_pairs_clear(Q_rank_cutoff=0.2)
        restot_red = fitter_red.residuals_average

        if restot_last < restot:
            fitter_use = fitter_last
            if doc_db is not None:
                doc_db.update_recursive(doc_db_last)
            aid.log("Using last (direct)!", fitter_ord(fitter_last))
            break

        if restot_last < restot_red:
            fitter_use = fitter_last
            if doc_db is not None:
                doc_db.update_recursive(doc_db_last)
            aid.log("Using last (reduced)!", fitter_ord(fitter_last))
            break

        if restot_last < 1.10 * restot:
            ord_red = fitter_ord(fitter_red)
            ord_last = fitter_ord(fitter_last)
            if ord_red < ord_last:
                fitter_use = fitter_red
                if doc_db is not None:
                    doc_db.update_recursive(doc_db_current)
            else:
                fitter_use = fitter_last
                if doc_db is not None:
                    doc_db.update_recursive(doc_db_last)
            aid.log("Using current")
            break
        # else continue the loop
        fitter_last = fitter
        restot_last = restot
        doc_db_last = doc_db_current

    fitter = fitter_use
    if nyquist_final_Hz == "keep":
        # this means to not perform the nyquist shift,
        # only really used in testing
        pass
    elif nyquist_final_Hz is None or nyquist_final_Hz == 0:
        ZPKrep = nyquist_move.nyquist_remove(
            fitter,
            split_neg_real=True,
            clear_neg_real=True,
            aid=aid,
        )
        fitter = ChebychevFilter(
            ZPKrep=ZPKrep,
        )
    elif nyquist_final_Hz > 0:
        fitter = nyquist_move.nyquist_move(
            fitter,
            nyquist_new=nyquist_final_Hz,
            split_neg_real=True,
            clear_neg_real=True,
            aid=aid,
        )
    else:
        raise RuntimeError(
            (
                "Nyquist cannot be negative. "
                " Positive for Z-domain, None or 0 for S."
                ' or "keep" to maintain the maxdata nyquist'
            )
        )

    return fitter


def ratdisc_single(
    argB=args.UNSPEC,
    F_Hz=args.UNSPEC,
    data=args.UNSPEC,
    nyquist_final_Hz=args.UNSPEC,
    SNR=args.UNSPEC,
    SNR_cutoff=args.UNSPEC,
    ZPK_overlay=args.UNSPEC,
    order=args.UNSPEC,
    complete=args.UNSPEC,
    doc_db=None,
    aid=None,
    no_nyquist_shift=False,
):
    aid = ensure_aid(aid)

    F_Hz = args.argscan(locals(), argB, args.REQ, arg="F_Hz")
    data = args.argscan(locals(), argB, args.REQ, arg="data")
    nyquist_final_Hz = args.argscan(locals(), argB, None, arg="nyquist_final_Hz")
    SNR = args.argscan(locals(), argB, args.REQ, arg="SNR")
    SNR_cutoff = args.argscan(locals(), argB, 100, arg="SNR_cutoff")
    ZPK_overlay = args.argscan(locals(), argB, None, arg="ZPK_overlay")
    order = args.argscan(locals(), argB, 50, arg="order")
    complete = args.argscan(locals(), argB, True, arg="complete")

    if nyquist_final_Hz is not None and nyquist_final_Hz <= 0:
        raise RuntimeError("nyquist_final_Hz must be >0 or None")

    F_Hz_sort = np.sort(F_Hz)
    F_nyquist_Hz = F_Hz_sort[-1] + (F_Hz_sort[-1] - F_Hz_sort[-2])
    F_nyquist_Hz = 1 * F_Hz_sort[-1]

    if SNR_cutoff is not None:
        if SNR is None:
            SNR = 1
        else:
            SNR = np.minimum(SNR, SNR_cutoff)

    kwargs = dict(
        # delay_s = -.5 / F_nyquist_Hz
    )

    if ZPK_overlay is not None:
        ZPK_overlay = representations.asZPKTF(
            ZPK_overlay,
            complete=complete,
        )
        ZPK_overlay_rep = TFmath.SorZtoSorZ(
            ZPK_overlay,
            F_nyquist_Hz_in=nyquist_final_Hz,
            F_nyquist_Hz_out=F_nyquist_Hz,
            error=True,
        )
        Zov, Pov, Kstart = ZPK_overlay_rep
    else:
        Zov = ()
        Pov = ()
        Kstart = 1

    fitter_x = RationalDiscFilter(
        F_Hz=F_Hz,
        data=data,
        W=SNR,
        npoles=order,
        nzeros=order,
        F_nyquist_Hz=F_nyquist_Hz,
        zeros_overlay=Zov,
        poles_overlay=Pov,
        gain=Kstart,
        **kwargs
    )
    fitter_x.mindelay = True
    fitter_x.stabilize = True
    fitter_x.fit_zeros()
    fitter_x.fit_poles()
    fitter_x.mindelay = False
    fitter_x.fit_zeros()
    fitter_x.fit_poles()
    annotate.annotate(
        doc_db,
        name="initial_direct",
        fitter=fitter_x,
        plotter=plots.plot_fitter_flag,
        method="fit_poles, fit_zeros",
        verbosity=9,
        about=(
            """
            initial guess without SVD technique
            """
        ),
    )

    fitter_p = RationalDiscFilter(
        # parent       = fitter_x,
        F_Hz=F_Hz,
        data=data,
        W=SNR,
        npoles=order,
        nzeros=order,
        F_nyquist_Hz=F_nyquist_Hz,
        zeros_overlay=Zov,
        poles_overlay=Pov,
        gain=Kstart,
        **kwargs
    )
    fitter_p.mindelay = True
    fitter_p.stabilize = True
    fitter_p.fit_poles_mod_zeros()
    fitter_p.fit_zeros()
    fitter_p.mindelay = False
    fitter_p.stabilize = True
    fitter_p.fit_poles()
    fitter_p.fit_zeros()
    fitter_p.fit_poles()
    fitter_p.fit_zeros()
    fitter_p.fit_poles()
    annotate.annotate(
        doc_db,
        name="initial_poles",
        fitter=fitter_p,
        plotter=plots.plot_fitter_flag,
        method="fit_poles_mod_zeros",
        verbosity=9,
        about=(
            """
            Performs the SVD for a rough initial guess
            """
        ),
    )

    #
    rep_z = RationalDiscFilter(
        # parent       = fitter_x,
        F_Hz=F_Hz,
        data=data,
        W=SNR,
        npoles=order,
        nzeros=order,
        F_nyquist_Hz=F_nyquist_Hz,
        zeros_overlay=Zov,
        poles_overlay=Pov,
        gain=Kstart,
        **kwargs
    )
    # put on mindelay for zeros to help ensure stable poles in froissart doublets
    rep_z.mindelay = True
    rep_z.stabilize = False
    rep_z.fit_zeros_mod_poles()
    rep_z.fit_poles()
    rep_z.mindelay = False
    rep_z.stabilize = True
    rep_z.fit_zeros()
    rep_z.fit_poles()
    rep_z.fit_zeros()
    rep_z.fit_poles()
    annotate.annotate(
        doc_db,
        name="initial_zeros",
        fitter=rep_z,
        plotter=plots.plot_fitter_flag,
        method="fit_poles_mod_zeros",
        verbosity=9,
        about=(
            """
            Performs the SVD for a rough initial guess
            """
        ),
    )

    aid.log(
        "(direct = {0:.3e}, Psvd= {0:.3e}, Zsvd= {0:.3e})".format(
            fitter_x.residuals_average,
            fitter_p.residuals_average,
            rep_z.residuals_average,
        )
    )
    fitter_list = [
        fitter_x,
        fitter_p,
        rep_z,
    ]
    fitter_list.sort(key=lambda f: f.residuals_average)
    fitter = fitter_list[0]
    if fitter is fitter_p:
        annotate.annotate(
            doc_db,
            name="choose poles",
            method="if",
            verbosity=3,
            about=(
                """
                Chose the Poles SVD fitter as it had the smaller residual of {0:.2e} vs. {1:.2e} for the zeros
                """.format(
                    fitter_p.residuals_average, rep_z.residuals_average
                )
            ),
        )
        # fitter.fit_poles_mod_zeros()
        # fitter.fit_zeros()
    elif fitter is rep_z:
        annotate.annotate(
            doc_db,
            name="choose zeros",
            method="if",
            verbosity=3,
            about=(
                """
                Chose the zeros SVD fitter as it had the smaller residual of {0:.2e} vs. {1:.2e} for the poles
                """.format(
                    rep_z.residuals_average, fitter_p.residuals_average
                )
            ),
        )
        # fitter.fit_zeros_mod_poles()
        # fitter.fit_poles()
    elif fitter is fitter_x:
        annotate.annotate(
            doc_db,
            name="choose direct",
            method="if",
            verbosity=3,
            about=(
                """
                Chose the direct (non-SVD) fitter as it had the smallest residual of {0:.2e}
                """.format(
                    fitter_x.residuals_average
                )
            ),
        )
    fitter.match_pair_iter(Q_rank_cutoff=0.4, zeros_first=False)
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.match_pair_iter(Q_rank_cutoff=0.4)
    fitter.fit_poles()
    fitter.fit_zeros()

    fitter.mindelay = False
    fitter.stabilize = True

    r_last = 1e10
    for idx in range(3):
        fitter.match_pair_iter(Q_rank_cutoff=0.4)
        fitter.fit_poles()
        r_p = fitter.residuals_average
        fitter.fit_zeros()
        r_z = fitter.residuals_average
        r = min(r_p, r_z)
        # aid.log(r / r_last)
        r_last = r

    fitter.stabilize = True

    annotate.annotate(
        doc_db,
        name="seq_iter_3",
        fitter=fitter,
        plotter=plots.plot_fitter_flag,
        method="RationalDiscFilter.fit_poles, RationalDiscFilter.fit_zeros",
        about=(
            """
            First iterations, enforcing stabilized poles residual of {0:.2e}
            """.format(
                fitter.residuals_average
            )
        ),
        verbosity=9,
    )

    fitter.fit_zeros()
    r_z = fitter.residuals_average
    fitter.fit_poles()
    r_p = fitter.residuals_average
    if r_z < r_p:
        # FIT POLES SHOULD BE THE LAST FIT! The residuals tend to be lowest under the poles
        fitter.fit_zeros()
        pass
    ##fitter.stabilize = False
    ##fitter.fit_poles()
    # fitter.fit_zeros()

    annotate.annotate(
        doc_db,
        name="seq_iter_4",
        fitter=fitter,
        plotter=plots.plot_fitter_flag,
        method="RationalDiscFilter.fit_poles, RationalDiscFilter.fit_zeros",
        about=(
            """
            First iterations, enforcing stabilized poles residual of {0:.2e}
            """.format(
                fitter.residuals_average
            )
        ),
        verbosity=4,
    )
    aid.log("LINEAR Final Residuals: ", fitter.residuals_average)

    if no_nyquist_shift:
        # this means to not perform the nyquist shift,
        # used in the full disc sequence loop
        pass
    elif nyquist_final_Hz is None or nyquist_final_Hz == 0:
        ZPKrep = nyquist_move.nyquist_remove(
            fitter,
            split_neg_real=True,
            clear_neg_real=True,
            aid=aid,
        )
        fitter = ChebychevFilter(
            ZPKrep=ZPKrep,
        )
    elif nyquist_final_Hz > 0:
        fitter = nyquist_move.nyquist_move(
            fitter,
            nyquist_new=nyquist_final_Hz,
            split_neg_real=True,
            clear_neg_real=True,
            aid=aid,
        )
    else:
        raise RuntimeError(
            (
                "Nyquist cannot be negative. "
                " Positive for Z-domain, None or 0 for S."
                ' or "keep" to maintain the maxdata nyquist'
            )
        )

    # annotate.annotate(
    #    doc_db,
    #    name     = 'nyquist_move',
    #    fitter   = fitter,
    #    method   = 'RationalDiscFilter.nyquist_move',
    #    plotter  = plots.plot_fitter_flag,
    #    about    = (
    #        """
    #        Moves to intended nyquist, clipping negative real poles/zeros likely caused by phase defects of low nyquist
    #        """.format(fitter.residuals_average)
    #    ),
    #    verbosity       = 3,
    # )

    # NOTE: shouldn't currently calculate residuals_average using the post-nyquist fitter as it is not numerically stable
    # TODO: switch to MRF for this residual
    annotate.annotate(
        doc_db,
        name="Final",
        method="ratdisc_single",
        fitter=fitter,
        plotter=plots.plot_fitter_flag,
        about=(
            """
            Uses SVD to create initial guess of fit for data, followed by several iterative fits
            (see reference ???).

               * It is a linear method, finding global optimum (nonlocal). This makes it get stuck if systematics are bad. To prevent this,
            it requires gratuitous overfitting to reliably get good fits.
               * It requires a nyquist frequency that is very low, near the last data point. This can cause artifacts due to phasing discontinuity near the nyquist.
               * The provided nyquist frequency is shifted up at the end, removing the real poles/zeros that are typically due to phasing discontinuity

            """.format()
        ),
        verbosity=2,
    )
    return fitter
